import curses
import logging
import requests
import os
import sys

from datetime import datetime
from clustersitter import MonitoredMachine, ProductionJob
from machineconsole.main import (
    MachineManagementScreen
)
from machineconsole.menu import MenuChanger, MenuOption, Table
from sittercommon import arg_parser
from sittercommon.api import ClusterState
from sittercommon.machinedata import MachineData
from sittercommon.utils import load_defaults, write_defaults, strip_html


class ClusterManagementScreen(MachineManagementScreen):
    def __init__(self, cluster_data, *args, **kwargs):
        super(MachineManagementScreen, self).__init__(*args, **kwargs)
        self.cluster_data = cluster_data
        self.machine_data = None

    def header(self):
        self.add_line("#" * self.scr.getmaxyx()[1])
        self.add_line(
            "ClusterSitter at %s - Curses UI %s" % (
                self.cluster_data.url, datetime.now()))
        self.add_line("#" * self.scr.getmaxyx()[1])

    def reload_data(self):
        if self.machine_data:
            self.machine_data.reload()
        else:
            self.cluster_data.reload()

    def basic_tasks(self):
        hotkey = 1
        table = None
        if self.current_loc == "mainmenu":
            self.machine_data = None
            self.reload_data()
            table = Table(self.scr.getmaxyx()[1],
                          ["Running Job Name", "Active Instances",
                           "Command"])
            for job in self.cluster_data.jobs:
                table.add_row(
                    [job.name,
                     sum([len(v) for v in job.fill_machines.values()]),
                     job.task_configuration['command']])

                option = MenuOption(
                    job.name,
                    action=MenuChanger(self.change_menu, "show_job",
                                       (job.name, job)),
                    hotkey=str(hotkey),
                    hidden=True)
                hotkey += 1
                self.factory.add_default_option(option)
        else:
            if len(self.aux) == 2 and isinstance(self.aux[1], ProductionJob):
                table = Table(self.scr.getmaxyx()[1],
                              ["Machine Name", "Machine IP",
                               "Machine Config"])
                job = self.aux[1]
                for machine in self.cluster_data.get_machines_for_job(job):
                    table.add_row(
                        [machine.hostname, machine.config.ip, machine.config])

                    option = MenuOption(
                        machine.hostname,
                        action=MenuChanger(self.change_menu, "show_machine",
                                           (machine.hostname, machine)),
                        hotkey=str(hotkey),
                        hidden=True)
                    hotkey += 1
                    self.factory.add_default_option(option)

            elif len(self.aux) == 2 and isinstance(
                    self.aux[1], MonitoredMachine):
                self.setup_machine()

        if self.machine_data:
            super(ClusterManagementScreen, self).basic_tasks()

        if table:
            table.render(self)

    def mainmenu(self):
        menu = self.factory.new_menu("Main Menu")
        menu.add_option_vals("Refresh Window", action=dir, hotkey="*")
        menu.add_option_vals("Add a new Job",
                             action=lambda: self.change_menu('addtask'))

        menu.render()

    def show_job(self):
        name, job = self.aux

        menu = self.factory.new_menu(
            "%s (%s)" % (
                name,
                job.task_configuration['command'],
            ))

        menu.add_option_vals(
            "Main Menu",
            action=lambda: self.change_menu(
                'mainmenu'), hotkey="*")

        menu.add_option_vals("Pause Job",
                             action=lambda: None)

        menu.add_option_vals("Add New Instance",
                             action=lambda: None)

        menu.add_option_vals("Remove an Instance",
                             action=lambda: None)

        menu.render()

    def setup_machine(self):
        hostname, machine = self.aux
        self.machine_data = MachineData(hostname, 40000)

    def show_machine(self):
        super(ClusterManagementScreen, self).mainmenu()

    def show_log(self, task):
        curses.endwin()
        os.system("clear")
        stdout = strip_html("%s/logfile?logname=%s.%s&tail=100" % (
            task['monitoring'],
            'stdout',
            task['num_task_starts'] - 1))
        print requests.get(stdout).content
        raw_input()


def main(sys_args=None):
    if not sys_args:
        sys_args = sys.argv[1:]

    default_options = load_defaults()
    parser = arg_parser.ArgumentParser(
        description="Cerebro Command Line Interaction Tool")

    parser.add_argument(
        "--clustersitterurl", dest="clustersitter_url",
        help="URL to the root of the clustersitter",
        default=default_options.get('clustersitter_url'))

    args = parser.parse_args(sys_args)
    default_options['clustersitter_url'] = args.clustersitter_url
    write_defaults(default_options)

    # Silence misc. logging.getLogger() 'can't find handler' warnings
    logging.basicConfig()
    logging.getLogger().setLevel(logging.CRITICAL)

    cluster_data = ClusterState(default_options['clustersitter_url'])

    screen = ClusterManagementScreen(cluster_data)

    try:
        if not cluster_data.url:
            print "Couldn't find a running cluster sitter!"
            return

        screen.run()
    except:
        curses.endwin()
        import traceback
        traceback.print_exc()

if __name__ == '__main__':
    main()



import json
import requests

from sittercommon import arg_parser


def get_help_string():
    return "Update a job to a new version"


def get_command():
    return "updatejob"


def get_parser(parser):
    parser.add_argument(dest="job_name",
                        help='The name of the file to upload in'
                        ' the appropriate releases folder.  Leave blank for '
                        'latest.')

    parser.add_argument("--version", dest="version",
                        help='The name of the file (without .tgz) to upload in'
                        ' the appropriate releases folder.  Leave blank for '
                        'latest.')

    return parser


def run_command(clustersitter_url=None,
                job_name=None,
                version=None):

    print "%s/update_job" % clustersitter_url
    data = {'data': json.dumps({'job_name': job_name,
                                'version': version})}
    resp = requests.post("%s/update_job" % clustersitter_url, data=data)
    print resp.content



import json
import simplejson
import subprocess
import time


class TaskManager(object):
    required_fields = ['name', 'command']
    optional_fields = [
        'auto_start',
        'restart',
        'max_restarts',
        'ensure_alive',
        'poll_interval',
        'allow_exit',
        'cpu',
        'mem',
        'time_limit',
        'uid']

    def __init__(self, task_definition, log_location, launch_location):
        self.reload_from_definition(task_definition)
        self.launch_location = launch_location

        self.http_monitoring = True
        self.http_monitoring_port = None
        self.stdout_location = log_location
        self.stderr_location = log_location

        self.was_started = False

        self.sitter_stdout = "%s/%s.stdout" % (log_location, self.name)
        self.sitter_stderr = "%s/%s.stderr" % (log_location, self.name)

        # Pre-clear stdout/stderr files
        open(self.sitter_stdout, 'w').close()
        open(self.sitter_stderr, 'w').close()

        self.process = None
        self.used_pids = []

    def reload_from_definition(self, task_definition):
        self.auto_start = task_definition.get('auto_start', False)
        self.restart = task_definition.get('restart', False)
        self.max_restarts = task_definition.get('max_restarts', -1)
        self.ensure_alive = task_definition.get('ensure_alive', False)
        self.poll_interval = task_definition.get('poll_interval', 0.1)
        self.allow_exit = task_definition.get('allow_exit', False)
        self.cpu = task_definition.get('cpu')
        self.mem = task_definition.get('mem')
        self.time_limit = task_definition.get('time_limit')
        self.uid = task_definition.get('uid')
        self.command = task_definition['command']
        self.name = task_definition['name']

    def __repr__(self):
        return json.dumps(self.to_dict())

    def to_dict(self):
        data = {}
        for opt in self.required_fields:
            data[opt] = getattr(self, opt)

        for opt in self.optional_fields:
            val = getattr(self, opt)
            if val is not None:
                data[opt] = val

        return data

    def set_port(self, port):
        self.http_monitoring_port = port

    def initialize(self):
        if self.auto_start:
            self.start()

    def stdall(self):
        stdout = open(self.sitter_stdout).read()
        stderr = open(self.sitter_stderr).read()
        return stdout, stderr

    def get_old_logfilenames(self):
        stdout = open(self.sitter_stdout).readlines()
        print stdout
        try:
            return simplejson.loads(stdout[-1])
        except:
            return {}

    def is_running(self):
        if not self.process:
            return False

        return self.process.poll() == None

    def get_last_pid(self):
        return self.used_pids[-1]

    def stop(self):
        print "Stopping %s" % self.name
        self.was_started = False

        while self.is_running():
            if self.process:
                self.process.terminate()
            time.sleep(.1)

    def start(self):
        args = ["tasksitter"]

        if self.restart:
            args.append("--restart")

        if self.max_restarts:
            args.append("--max-restarts=%s" % self.max_restarts)

        if self.ensure_alive:
            args.append("--ensure-alive")

        if self.poll_interval:
            args.append("--poll-interval=%s" % self.poll_interval)

        if self.cpu:
            args.append("--cpu=%s" % self.cpu)

        if self.mem:
            args.append("--mem=%s" % self.mem)

        if self.time_limit:
            args.append("--time-limit=%s" % self.time_limit)

        if self.uid:
            args.append("--uid=%s" % self.uid)

        args.append("--command")
        args.append(self.command)

        args.append("--http-monitoring")
        args.append("--http-monitoring-port=%s" % self.http_monitoring_port)

        args.append("--stdout-location=%s" % self.stdout_location)
        args.append("--stderr-location=%s" % self.stderr_location)

        print "Executing command %s" % args

        self.process = subprocess.Popen(
            args,
            stdout=open(self.sitter_stdout, 'a'),
            stderr=open(self.sitter_stderr, 'a'))

        self.was_started = True
        self.used_pids.append(self.process.pid)

        print "Task '%s' Started.  Task Monitoring At: http://localhost:%s" % (
            self.command,
            self.http_monitoring_port)



import logging
import os
import socket
import sys
import threading
from datetime import datetime
from logging import FileHandler

import actions
import clusterstate
import deploymentrecipe
import dynect
import eventmanager
import jobfiller
import machinemonitor
import monitoredmachine
import productionjob
import providers.aws
import sittercommon.machinedata

from clusterstate import ClusterState
from clusterstats import ClusterStats
from eventmanager import ClusterEventManager
from machinemonitor import MachineMonitor
from monitoredmachine import MonitoredMachine
from productionjob import ProductionJob
from providers.aws import AmazonEC2
from sittercommon import http_monitor
from sittercommon import logmanager

logger = logging.getLogger(__name__)

"""
In an ideal world this is written such that we always speak
in terms of CPUs and RAM requirements for each job.  This allows
us to put multiple jobs on a machine whilst keeping track
of total theoretical machine utilization.  However, for simplicities
sake we're cutting some corners (FOR NOW) and assuming one job
per machine.  These sections are noted with the comment
#!MACHINEASSUMPTION!
"""


class ClusterSitter(object):
    def __init__(self, log_location, daemon,
                 provider_config,
                 dns_provider_config,
                 keys=None, login_user=None,
                 starting_port=30000,
                 launch_location=None):
        self.worker_thread_count = 4
        self.daemon = daemon
        self.keys = keys
        self.login_user = login_user
        self.provider_config = provider_config
        self.dns_provider_config = dns_provider_config
        self.log_location = log_location
        self.launch_location = launch_location
        self.launch_time = datetime.now()
        self.start_state = "Not Started"

        # In seconds
        self.stats_poll_interval = 5

        self.state = ClusterState(self)

        self.orig_starting_port = starting_port
        self.next_port = starting_port
        self.start_count = 1
        self.command = "clustersitter"
        self.parent_pid = os.getpid()
        self.logmanager = logmanager.LogManager(stdout_location=log_location,
                                                stderr_location=log_location)
        self.logmanager.set_harness(self)

        self.dns_provider = None
        if 'class' in dns_provider_config:
            (module, clsname) = dns_provider_config['class'].split(':')
            clz = self._get_recipe_class(module, find_func=clsname)
            self.dns_provider = getattr(clz, clsname)(
                dns_provider_config)

        self.stats = ClusterStats(self)
        self.http_monitor = http_monitor.HTTPMonitor(self.stats,
                                                     self,
                                                     30000)

        self.http_monitor.add_handler('/overview', self.stats.overview)
        self.http_monitor.add_handler('/add_job', self.api_add_job)
        self.http_monitor.add_handler('/remove_job', self.api_remove_job)
        self.http_monitor.add_handler('/update_idle_limit',
                                      self.api_enforce_idle)
        self.http_monitor.add_handler('/update_logging_level',
                                      self.api_update_logging_level)
        self.http_monitor.add_handler('/update_job',
                                      self.api_update_job)

        # Do lots of logging configuration
        modules = [
            sys.modules[__name__],
            actions,
            clusterstate,
            deploymentrecipe,
            dynect,
            eventmanager,
            jobfiller,
            machinemonitor,
            monitoredmachine,
            productionjob,
            providers.aws,
            sittercommon.machinedata,
        ]

        formatter = logging.Formatter(
            '%(asctime)s - %(name)s:%(threadName)s'
            '.%(levelname)s - %(message)s')

        if not os.path.exists(self.log_location):
            os.makedirs(self.log_location)

        all_file = FileHandler("%s/all.log" % self.log_location)
        all_file.setFormatter(formatter)

        self.logfiles = ["%s/all.log" % self.log_location]

        for module in modules:
            name = module.__name__.split('.')[-1]
            logfile = "%s/%s.log" % (
                self.log_location, name)
            self.logfiles.append(logfile)
            self.logmanager.add_logfile(name, logfile)
            handler = FileHandler(logfile)
            handler.setFormatter(formatter)
            module.logger.addHandler(all_file)
            module.logger.addHandler(handler)
            self.state.loggers.append(module.logger)

        requests_logger = logging.getLogger(
            'requests.packages.urllib3.connectionpool')
        requests_logger.setLevel(logging.WARNING)

        socket.setdefaulttimeout(2)

    # ----------- API ----------------
    def _api_check(self, args, required_fields):
        if self.start_state != "Started":
            return "Not ready to recieve API calls"

        for field in required_fields:
            if not field in args:
                return "Missing Field %s" % field

        return

    def api_update_logging_level(self, args):
        # Default level is INFO
        level = args.get('level', 20)
        for logger in self.state.loggers:
            logger.setLevel(level)

        ClusterEventManager.handle(
            "Updated logging level to %s" % level)

        return "Level set to %s" % level

    def api_enforce_idle(self, args):
        # Really naive right now, a global # of
        # max idle per zone.  Could do a lot more here.
        check = self._api_check(args, ['idle_count_per_zone'])

        if check:
            return check

        try:
            self.state.max_idle_per_zone = int(args['idle_count_per_zone'])
        except:
            return "Invalid limit"

        ClusterEventManager.handle(
            "Enforce Idle Limit at %s" % int(args['idle_count_per_zone']))
        return "Limit set"

    def api_update_job(self, args):
        check = self._api_check(args,
                                ['job_name'])

        if check:
            return check

        job_name = args['job_name']
        if not self.state.update_job(job_name):
            return "Error updating job: %s doesn't exist" % job_name

        ClusterEventManager.handle(
            'Update %s started' % job_name)
        return "Job update initiated"

    def api_add_job(self, args):
        check = self._api_check(args,
                                ['dns_basename',
                                 'task_configuration',
                                 'deployment_layout',
                                 'deployment_recipe',
                                 'recipe_options',
                                 'persistent'])
        if check:
            return check
        job = ProductionJob(
            self,
            args['dns_basename'],
            args['task_configuration'],
            args['deployment_layout'],
            args['deployment_recipe'],
            args['recipe_options'],
            args['persistent'],
            args.get('linked_job'))

        if args.get('linked_job'):
            job.find_linked_job()
            if not job.linked_job_object:
                return "Couldn't find linked job!"

        if self.state.add_job(job):
            ClusterEventManager.handle(
                "Added a job: %s" % job.get_name())

            return "Job Added"
        else:
            return "Error adding job, see logs"

    def api_remove_job(self, args):
        check = self._api_check(args, ['name'])

        if check:
            return check

        jobs = self.state.remove_job(args['name'])
        ClusterEventManager.handle(
            "Removed jobs: %s" % ', '.join(jobs))
        if jobs:
            return "Removed: %s" % ', '.join(jobs)
        else:
            return "Job Not Found"

    # ----------- END API ----------

    def machines_in_queue(self):
        #TODO: Replace with an equivelent method in ClusterState.
        for monitor, thread in self.state.monitors:
            if monitor.processing_new_machines():
                return True

        return False

    def _get_recipe_class(self, recipe_class, find_func="run_deploy"):
        if isinstance(recipe_class, type):
            return recipe_class

        recipe_cls = None
        if isinstance(recipe_class, basestring):
            try:
                recipe_cls = __import__(
                    recipe_class, globals(), locals())
            except:
                # Odd?
                logger.warn("Unable to load recipe %s..." % recipe_class)
        else:
            logger.warn("Invalid recipe class %s..." % recipe_class)
            return None

        if hasattr(recipe_cls, find_func):
            return recipe_cls

        exports = dir(recipe_cls)
        for export in exports:
            if hasattr(getattr(recipe_cls, export), find_func):
                return getattr(recipe_cls, export)

        try:
            module = sys.modules[recipe_class]
        except:
            return None

        exports = dir(module)
        for export in exports:
            if hasattr(getattr(module, export), find_func):
                value = getattr(module, export)
                if value.__module__ == module.__name__:
                    return value

    def build_recipe(self, recipe_class, machine,
                     post_callback=None, options=None,
                     given_logger=None):
        username = self.login_user
        keys = self.keys
        if machine.config.login_name:
            username = machine.config.login_name

        if machine.config.login_key:
            keys.append(machine.config.login_key)

        recipe_cls = self._get_recipe_class(recipe_class)
        if not recipe_cls:
            return None

        return recipe_cls(machine.config.hostname,
                          username,
                          keys,
                          post_callback=post_callback,
                          options=options,
                          given_logger=given_logger,
                          dns_hostname=machine.config.dns_name,
                          launch_location=self.launch_location)

    def get_next_port(self):
        works = None
        while not works:
            try:
                sok = socket.socket(socket.AF_INET)
                sok.bind(("0.0.0.0", self.next_port))
                sok.close()
                works = self.next_port
                self.next_port += 1
            except:
                self.next_port += 1
                if self.next_port > self.orig_starting_port + 10000:
                    self.next_port = self.orig_starting_port
        return works

    def decomission_machine(self, machine):
        self.state.remove_machine(machine)
        provider = self.state.get_zone_provider(
            machine.config.shared_fate_zone)
        if not provider:
            logger.warn(
                "No provider found for %s?" % machine.config.shared_fate_zone)
            return

        ClusterEventManager.handle(
            "Decomissioning %s" % str(machine))

        if not provider.decomission(machine):
            # If we can't decomission it then perhaps its locked
            # and we should leave well enough alone at this point,
            # just remove it from monitoring etc.
            ClusterEventManager.handle(
                "Provider doesn't allow decomissioning of %s" % str(machine))
            return

        if machine.config.dns_name:
            self.dns_provider.remove_record(data=machine.config.ip,
                                            hostName=machine.config.dns_name)

            # Strip off the leading number, e.g.
            # 12.bar.mydomain.com -> bar.mydomain.com
            root_name = '.'.join(machine.config.dns_name.split('.')[1:])

            self.dns_provider.remove_record(data=machine.config.ip,
                                            hostName=root_name)

        # Now look for other dangling records pointing to this machine
        # and delete those too.
        records = self.dns_provider.get_records()
        for record in records:
            if record['value'] == machine.config.ip:
                logger.info(
                    "Removing %s from %s" % (
                    machine.config.ip, record['record']))
                self.dns_provider.remove_record(data=machine.config.ip,
                                                hostName=record['record'])

        ClusterEventManager.handle(
            "Decomissioning of %s complete!" % str(machine))

    def add_machines(self, machines, update_dns=True, deploying=False):
        """
        Add machines to the sitter.

        @param machines A list of machines to add.
        @param update_dns Whethor or not to update the DNS for the machines.
            Defaults to True.
        @param deploying Whether or not to flag the machines as currently being
            deployed. Defaults to False.
        """
        if not machines:
            return

        monitored_machines = []
        logger.info("Adding machines to monitoring: %s" % machines)
        for m in machines:
            mm = m
            if not isinstance(mm, MonitoredMachine):
                mm = MonitoredMachine(m)
            status = None
            if deploying:
                status = self.state.Deploying
            self.state.monitor_machine(mm)
            self.state.add_machine(mm, status=status, existing=True)
            monitored_machines.append(mm)

        if not update_dns:
            return

        # Ensure we have up to date DNS names for each machine
        logger.info("Loading DNS Records...")

        records = self.dns_provider.get_records()
        record_by_ip = {}
        for record in records:
            record_by_ip[record['value']] = record['record']

        for machine in monitored_machines:
            if not machine.config.dns_name:
                ip = socket.gethostbyname(machine.config.hostname)
                if ip in record_by_ip:
                    machine.config.dns_name = record_by_ip[ip]
                    logger.info(
                        "Found name %s for %s" % (
                        machine.config.dns_name, ip))

    def start(self):
        self.http_monitor.start()
        logger.info((
            "Cluster Sitter Monitor started at "
            "http://localhost:%s") % self.http_monitor.port)

        self.start_state = "Starting Up"

        logger.info("Initializing MachineProviders")
        # Initialize (non-started) machine monitors first
        # Then download machine data, populate data structure and monitors
        # then kickoff everything.

        logger.info(
            "Initializing %s MachineMonitors" % self.worker_thread_count)
        # Spin up all the monitoring threads
        #TODO: Move monitor thread spinup to ClusterState.
        for threadnum in range(self.worker_thread_count):
            machinemonitor = MachineMonitor(parent=self,
                                            number=threadnum)
            thread = threading.Thread(target=machinemonitor.start,
                                      name='Monitoring-%s' % threadnum)
            self.state.monitors.append((machinemonitor, thread))

        #TODO: Don't hard code providers.
        aws = AmazonEC2(self.provider_config['aws'])
        if aws.usable():
            self.state.add_provider('aws', aws)

        for provider in self.state.get_providers().values():
            # Note: add_machines() has to be called AFTER the monitors
            # are initialized.
            logger.info(
                "adding %d machines for provider %s" % (
                len(provider.get_machine_list()), 'aws'))
            self.add_machines(provider.get_machine_list())

        logger.info("Zone List: %s" % self.state.get_zones())

        # Kickoff all the threads at once
        if self.daemon:
            self.logmanager.setup_all()

        logger.info(
            "Spinning up %s monitoring threads" %
            self.worker_thread_count)
        for monitor in self.state.monitors:
            monitor[1].start()

        logger.info("Starting metadata calculator")
        self.state.start()
        self.start_state = "Started"



from django.db import models
from djangotoolbox.fields import DictField


class DataRecord(models.Model):
    class Meta:
        app_label = 'clustersitter'

    timestamp = models.DateTimeField(auto_now_add=True)
    taskname = models.CharField(max_length=1024)
    machinename = models.CharField(max_length=1024)
    data = DictField()
    metadata = DictField()



import logging
import time
from datetime import datetime

logger = logging.getLogger(__name__)


class MachineMonitor:
    def __init__(self, parent, number, monitored_machines=[]):
        self.clustersitter = parent
        self.number = number
        # Need to copy the array to deref it so its not shared
        # amoung threads
        self.monitored_machines = [m for m in monitored_machines]
        self.add_queue = []
        self.pull_failures = {}
        self.failure_threshold = 5

        logger.info(
            "Initialized a machine monitor for %s" %
            str(self.monitored_machines))

    def num_monitored_machines(self):
        return len(self.monitored_machines) + len(self.add_queue)

    def has_machine(self, monitored_machine):
        """
        Check if this monitor has the given machine.

        @param monitored_machine The machine to check for.
        @return True if the monitor has the machine or False.
        """
        return (
            monitored_machine in self.monitored_machines or
            monitored_machine in self.add_queue)

    def remove_machine(self, monitored_machine):
        """
        Remove a monitored machine.

        @param monitored_machine The machine to remove.
        @return True if the machine was removed or False if the machine was not
            being monitored.
        """
        if monitored_machine in self.add_queue:
            self.add_queue.remove(monitored_machine)
            return True
        if monitored_machine in self.monitored_machines:
            self.monitored_machines.remove(monitored_machine)
            if monitored_machine in self.pull_failures:
                del self.pull_failures[monitored_machine]
            return True
        return False

    def add_machine(self, monitored_machine):
        """
        Add a machine to the monitor.

        @param monitored_machine The machine to add to the monitor.
        @return True if the machine is added to monitoring or False if the
            machine is already being monitored.
        """
        if self.has_machine(monitored_machine):
            return False

        self.add_queue.append(monitored_machine)
        self.pull_failures[monitored_machine] = 0

        logger.info(
            "Queued %s for inclusion in next stats run in %s" %
            (monitored_machine, self.number))
        return True

    def initialize_machines(self, monitored_machines):
        for m in monitored_machines:
            val = True
            try:
                val = m.initialize()
            except:
                val = False
                import traceback
                traceback.print_exc()
                logger.error(traceback.format_exc())
            if not val:
                self.pull_failures[m] += 1

    def __repr__(self):
        return str(self)

    def __str__(self):
        return ",".join([str(m) for m in self.monitored_machines])

    def processing_new_machines(self):
        if self.add_queue:
            return True

        for machine in self.monitored_machines:
            if not machine.has_loaded_data():
                return True

        return False

    def start(self):
        self.initialize_machines(self.monitored_machines)

        while True:
            start_time = datetime.now()
            try:
                logger.debug(
                    "Processing add queue %s at %s" % (
                    self.add_queue, self.number))
                while len(self.add_queue) > 0:
                    machine = self.add_queue[-1]
                    self.initialize_machines([machine])
                    self.monitored_machines.append(machine)
                    self.add_queue.remove(machine)

                logger.debug("Finished processing add queue")
                logger.debug(
                    "Beggining machine monitoring poll for %s at %s" % (
                        [str(a) for a in self.monitored_machines],
                        self.number))

                for machine in self.monitored_machines:
                    if machine.is_initialized():
                        val = True
                        try:
                            val = machine._api_get_stats()
                        except:
                            val = False
                            import traceback
                            traceback.print_exc()
                            logger.error(traceback.format_exc())

                        if not val:
                            self.pull_failures[machine] += 1
                            logger.info(
                                "Detected a pull failure for %s, total: %s" % (
                                    machine, self.pull_failures[machine]))

                        else:
                            self.pull_failures[machine] = 0
                    else:
                        self.initialize_machines([machine])

                logger.debug("Pull Failures: %s" % ([
                    (m.hostname, count) for m, count in
                    self.pull_failures.items()]))

                for machine, count in self.pull_failures.items():
                    if count >= self.failure_threshold:
                        self.monitored_machines.remove(machine)
                        del self.pull_failures[machine]
                        machine.detected_sitter_failures += 1
                        logger.warn((
                            "Removing '%s' because we can't contact "
                            "the sitter!") % machine.hostname)
            except:
                import traceback
                traceback.print_exc()
                logger.error(traceback.format_exc())

            time_spent = datetime.now() - start_time
            sleep_time = self.clustersitter.stats_poll_interval - \
                time_spent.seconds
            logger.debug(
                "Finished poll run for %s.  Time_spent: %s, sleep_time: %s" % (
                    [str(a) for a in self.monitored_machines],
                    time_spent,
                    sleep_time))

            if sleep_time > 0:
                time.sleep(sleep_time)



class MachineConfig(object):
    def __init__(self, hostname, shared_fate_zone,
                 cpus, mem, bits=64, disk=0, data=None,
                 ip=None):
        self.hostname = hostname
        self.cpus = cpus
        self.mem = mem
        self.shared_fate_zone = shared_fate_zone
        self.bits = bits
        self.disk = disk
        self.login_name = None
        self.login_key = None
        self.data = data
        self.dns_name = None
        self.ip = ip

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return "%s %sx %s-bit %sMBx%sGB" % (
            self.shared_fate_zone, self.cpus,
            self.bits, self.mem, self.disk)

    def serialize(self):
        return {
            'hostname': self.hostname,
            'cpus': self.cpus,
            'mem': self.mem,
            'shared_fate_zone': self.shared_fate_zone,
            'bits': self.bits,
            'disk': self.disk,
            'login_name': self.login_name,
            'login_key': self.login_key,
            'data': repr(self.data),
            'dns_name': self.dns_name,
            'ip': self.ip,
        }

    @classmethod
    def deserialize(cls, data):
        obj = cls(hostname=data['hostname'],
                  shared_fate_zone=data['shared_fate_zone'],
                  cpus=data['cpus'],
                  mem=data['mem'],
                  bits=data['bits'],
                  disk=data['disk'],
                  data=data['data'],
                  ip=data['ip'], )
        obj.login_name = data['login_name']
        obj.login_key = data['login_key']
        obj.dns_name = data['dns_name']

        return obj



import os
import requests
import simplejson
import sys
import zlib

from clustersitter.monitoredmachine import MonitoredMachine
from clustersitter.productionjob import ProductionJob


class ProviderConfig(object):
    def __init__(self):
        self.providers = {}

    def serialize(self):
        return self.providers

    @classmethod
    def deserialize(cls, data):
        obj = cls()
        obj.providers = data
        return obj

    def get_key_for_zone(self, shared_fate_zone):
        provider, zone = shared_fate_zone.split('-', 1)
        return self.providers.get(provider, {}).get(zone, {}).get('key_name')


class ClusterState(object):
    def __init__(self, url):
        self.url = url
        self.machines = []
        self.jobs = {}
        self.zones = {}
        self.provider_config = {}
        self.login_user = None
        self.raw = None

    def reload(self):
        url = "%s/overview?nohtml=1&format=json&compress=1" % self.url
        # 3 attempts, since sometimes downloading json is a bit flaky
        data = None
        for _ in xrange(3):
            response = requests.get(url)
            if response.status_code != 200:
                continue

            try:
                raw_data = zlib.decompress(response.content)
                data = simplejson.loads(raw_data)
                break
            except:
                continue

        if not data:
            sys.stderr.write("Couldn't load data!")
            sys.exit(1)

        self.raw = data
        self.jobs = [ProductionJob.deserialize(j) for j in data['jobs']]
        self.machines = [MonitoredMachine.deserialize(
            m) for m in data['machines']]
        self.provider_config = ProviderConfig.deserialize(
            data['provider_config'])

        self.keys = self.load_keys(data['keys'])
        self.login_user = data.get("login_user", data.get("username"))

    def find_key(self, key):
        return self.keys.get(key)

    def load_keys(self, keys):
        data = {}
        for key in keys:
            name = os.path.basename(key).replace('.pem', '')
            data[name] = key

        return data

    def get_job_names(self):
        return [j.name.lower() for j in self.jobs]

    def get_machines_for_job(self, job):
        machines = []
        for machine_list in job.fill_machines.values():
            for machine in machine_list:
                host, ip = machine.split(':')
                machines.append(self.get_machine(host))

        return machines

    def get_machine(self, hostname):
        for machine in self.machines:
            if hostname == machine.hostname:
                return machine

    def get_job(self, name):
        name = name.lower()
        for job in self.jobs:
            if job.name.lower() == name:
                return job

        return None



import tasksitter.main as main
import md5
import os
import sys
import tempfile
import unittest


class BasicTests(unittest.TestCase):

    def run_check(self, args, retval=0):
        try:
            main.main(args, allow_spam=True)
        except SystemExit, e:
            self.assertEqual(e.code, retval)

    def test_stdout_redirect(self):
        dirname = tempfile.mkdtemp()
        phrase = 'testing'
        command = "echo -n '%s'" % phrase
        self.run_check(["--command", command,
                        "--restart",
                        "--max-restarts=2",
                        "--ensure-alive",
                        "--stdout-location=%s" % dirname])
        data = []
        for i in range(0, 3):
            filename = "%s/%s.%s" % (
                dirname,
                md5.md5(command + str(os.getpid())).hexdigest(),
                i)

            data.append(open(filename).read())
            os.unlink(filename)

        os.rmdir(dirname)
        for d in data:
            self.assertEqual(d, phrase)

    def test_stderr_redirect(self):
        dirname = tempfile.mkdtemp()
        phrase = 'testing'
        command = "echo -n '%s' 1>&2" % phrase
        self.run_check(["--command", command,
                        "--restart",
                        "--max-restarts=2",
                        "--ensure-alive",
                        "--stderr-location=%s" % dirname])

        data = []
        for i in range(0, 3):
            filename = "%s/%s.%s" % (
                dirname,
                md5.md5(command + str(os.getpid()) + "err").hexdigest(),
                i)

            data.append(open(filename).read())
            os.unlink(filename)

        os.rmdir(dirname)
        for d in data:
            self.assertEqual(d, phrase)

    def test_quick_command(self):
        self.run_check(["--command", "ls >/dev/null"])

    def test_quick_command_cpu_constraint(self):
        self.run_check(["--cpu=.1", "--command", "ls >/dev/null"])

    def test_quick_command_mem_constraint(self):
        self.run_check(["--mem=10", "--command", "ls >/dev/null"])

    def test_quick_command_cpu_mem_constraint(self):
        self.run_check(["--cpu=.2", "--mem=10", "--command", "ls >/dev/null"])

    def test_slow_command(self):
        self.run_check(["--command", "ls >/dev/null; sleep .21"])

    # !FLAKY
    def _test_slow_command_cpu_constraint(self):
        self.run_check(["--cpu=.5", "--command", "ls >/dev/null; sleep .21"])

    def test_slow_command_mem_constraint(self):
        self.run_check(["--mem=10", "--command", "ls >/dev/null; sleep .21"])

 # FLAKY For some reason on slower machines
 #   def test_slow_command_cpu_mem_constraint(self):
 #       self.run_check(["--cpu=.9", "--mem=10",
 #                       "--command", "ls >/dev/null; sleep .31"])

    def test_time_constraint(self):
        self.run_check(["--time-limit=.1", "--command", "sleep .2"], 9)

    def test_time_constraint_inverse(self):
        self.run_check(["--time-limit=.5", "--command", "sleep .1"])

    def test_cpu_constraint(self):
        self.run_check(["--cpu=.5", "--command", "./spin.sh"], 9)

    def test_cpu_constraint_redirect(self):
        self.run_check(["--cpu=.1", "--command",
                        "bash -c './spin.sh 2>/dev/null'"],
                       9)

    def test_cpu_constraint_subprocess(self):
        self.run_check(["--cpu=.1", "--command",
                        "bash -c 'bash -c \'./spin.sh\''"],
                       9)

    def test_mem_constraint(self):
        self.run_check(["--mem=3", "--command", "./mem.sh"],
                       9)

    def test_mem_constraint_redirect_subprocess(self):
        self.run_check(["--mem=3", "--command",
                        "bash -c 'bash -c \'./mem.sh\''"],
                       9)

    def test_mem_constraint_redirect(self):
        self.run_check(["--mem=3",
                        "--command", "./mem.sh 2>/dev/null"], 9)

    def test_restarting(self):
        filename = tempfile.mktemp()
        self.run_check(["--cpu=.5", "--restart", "--max-restarts=2",
                        "--command",
                        "echo '1' >> %s ; ./spin.sh" % filename], 9)

        lines = open(filename).readlines()
        os.unlink(filename)
        self.assertEqual(3, len(lines))

    def test_ensure_alive(self):
        filename = tempfile.mktemp()
        self.run_check(["--ensure-alive", "--restart", "--max-restarts=2",
                        "--command",
                        "echo '1' >> %s" % filename])

        lines = open(filename).readlines()
        os.unlink(filename)
        self.assertEqual(3, len(lines))

    def test_ensure_alive_many_times(self):
        sys.setrecursionlimit(70)
        filename = tempfile.mktemp()
        self.run_check(["--ensure-alive", "--restart", "--max-restarts=71",
                        "--poll-interval=0.001",
                        "--command",
                        "echo '1' >> %s" % filename])

        sys.setrecursionlimit(500)

        lines = open(filename).readlines()
        os.unlink(filename)
        self.assertEqual(72, len(lines))



from sittercommon.api import ClusterState


def get_help_string():
    return "List jobs active in clustersitter"


def get_command():
    return "listjobs"


def get_parser(parser):
    return parser


def run_command(clustersitter_url=None):
    state = ClusterState(clustersitter_url)
    state.reload()
    for job in state.jobs:
        print "%s - %s instances" % (
            job.name,
            sum([len(v) for v in job.fill_machines.values()]))



"""
Cluster state tracking and maintenance.
"""

import logging
import simplejson
import time
from actions import (
    AddTaskAction, ClusterActionManager, DecomissionMachineAction,
    DeployMachineAction, RedeployMachineAction, RemoveTaskAction,
    RestartTaskAction, StartTaskAction, StopTaskAction)
from productionjob import ProductionJob
from threading import RLock, Thread

logger = logging.getLogger(__name__)


def lock(fn):
    """
    Lock the reentrant lock stored in self.lock during method execution.

    @param fn The method to lock.
    """
    def lock_wrapper(self, *args, **kwargs):
        self.lock.acquire()
        try:
            return fn(self, *args, **kwargs)
        finally:
            self.lock.release()
    return lock_wrapper


class JobState(object):
    """
    Job state. Manages the state of all running jobs on the cluster.

    Some static status attributes are defined as part of the class. These are:

    Deploying -- The task is being deployed.
    Running -- The task is running.
    Stopped -- The task is stopped.
    """

    Deploying = 'deploying'
    Running = 'running'
    Stopped = 'stopped'

    def __init__(self):
        """
        Initialize an empty state object.

        State attributes are structured as follows:

            tasks = {
                task_name: [
                    {
                        'machine': machine_object,
                        'status': task_status,
                        'zone': zone,
                    },
                ],
            }
        """
        self.tasks = {}

    def _get_task_item(self, name, machine):
        """
        Get the list item from self.tasks[name] that contains the given task.
        """
        if name in self.tasks:
            for task in self.tasks[name]:
                if task['machine'] == machine:
                    return task
        return None

    def flatten(self):
        """
        Flatten the job state. The result of the flattening is a set of
        three-tuples representing all of the tasks present in the state. Each
        tuple contains the zone, machine, and task name. Tasks that do not have
        machines assigned will not be included. This method is meant to ease
        comparisons of job state objects.

        @return The flattened job set.
        """
        flat = set()
        for name, task_items in self.tasks.iteritems():
            for task in task_items:
                if task['machine'] is not None:
                    flat.add((task['zone'], task['machine'], name))
        return flat

    def is_machine_idle(self, machine):
        """
        Check if a machine has any tasks assigned to it.

        @param machine The machine to check.
        @return True if the machine is idle or False.
        """
        for task_items in self.tasks.values():
            for task in task_items:
                if task['machine'] == machine:
                    return False
        return True

    def remove_machine(self, machine):
        """
        Remove tasks running on the given machine.

        @param machine The machine to remove tasks from.
        """
        for name in self.tasks.keys():
            for task in list(self.tasks[name]):
                if task['machine'] == machine:
                    self.tasks[name].remove(task)
            if not self.tasks[name]:
                del self.tasks[name]

    def has_task(self, name, machine=None):
        """
        Check if a task is deployed to the cluster.

        @param name The name of the task.
        @param machine Check if the task is deployed to this machine.
        @return True if the task is deployed or False.
        """
        if machine:
            return self._get_task_item(name, machine) is not None
        else:
            return name in self.tasks

    def is_task_deploying(self, name, machine):
        """
        Check if a task is being deployed.

        @param name The name of the task to check.
        @param machine The machine the task is being deployed to.
        @return True if the task is being deployed or False.
        """
        item = self._get_task_item(name, machine)
        return item and item['status'] == self.Deploying

    def get_task_status(self, name, machine):
        """
        Get the status of the task on the given machine.

        @param name The name of the task.
        @param machine The machine to check in.
        @return The status of the task or None if not present.
        """
        item = self._get_task_item(name, machine)
        if item:
            return item['status']
        return None

    def get_task_machines(self, name=None, zones=None):
        """
        Get machines that have tasks allocated to them.

        @param name Only return machines with this task running on them.
            Optional.
        @param zones Only return machines in these zones. Optional.
        @return A dictionary of zone/machine lists representing the found
            machines.
        """
        if zones:
            zoned_machines = {z: [] for z in zones}
        else:
            zoned_machines = {}

        if name is not None:
            if name not in self.tasks:
                task_items = {}
            else:
                task_items = {name: self.tasks[name]}
        else:
            task_items = self.tasks

        for name, tasks in task_items.iteritems():
            for task in tasks:
                zone = task['zone']
                machine = task['machine']
                if machine:
                    if zone not in zoned_machines:
                        zoned_machines[zone] = []
                    zoned_machines[zone].append(machine)
        return zoned_machines

    def get_machine_tasks(self, machine, status=None):
        """
        Get the tasks running on a machine.

        @param machine The machine to get the running tasks on.
        @param status Only get tasks matching this status. Optional.
        @return A set of task names running on the machine. An empty list if
            the machine is not present.
        """
        names = set()
        for name, tasks in self.tasks.iteritems():
            for task in tasks:
                if task['machine'] == machine:
                    if status is None or status == task['status']:
                        names.add(name)
        return names

    def add_tasks(self, name, zone, machines, create=None, status=None):
        """
        Add tasks to state tracking. The provided machines will be added if
        they are not present. More tasks than available machines may be added
        using the create parameter. Machines should be assigned to these tasks
        later using the set_pending_machines method.

        @param name The name of the task.
        @param zone The zone the tasks are being added to. Machines must be in
            this zone.
        @param machines The machines available to assign to the tasks.
        @param create A count of empty/undeployed machines to fill the task on.
        @param status The initial status of the tasks.
        """
        if create is None:
            create = 0
        if not status:
            status = self.Running

        if name not in self.tasks:
            self.tasks[name] = []

        for machine in machines:
            if not self.has_task(name, machine):
                self.tasks[name].append({
                    'machine': machine,
                    'status': status,
                    'zone': zone})
        for n in range(create):
            self.tasks[name].append({
                'machine': None,
                'status': status,
                'zone': zone})

    def remove_tasks(self, name, machines=None):
        """
        Remove tasks from machines.

        @param name The name of the tasks.
        @param machines The machines to remove the tasks from. Defaults to all
            machines.
        @return A list of tasks that were removed.
        """
        removed = []
        if name in self.tasks:
            removed.append(name)
            if machines is None:
                del self.tasks[name]
            else:
                for task in list(self.tasks[name]):
                    if task['machine'] in machines:
                        self.tasks[name].remove(task)
                if not self.tasks[name]:
                    del self.tasks[name]
        return removed

    def update_tasks(self, name, status, machines=None):
        """
        Update the status on a set of tasks.

        @param name The name of the task to update.
        @param machines The machines to update the task status on.
        """
        if name in self.tasks:
            for task in self.tasks[name]:
                if not machines or task['machine'] in machines:
                    task['status'] = status

    def set_pending_deploying(self, zone, name, count):
        """
        Flag a pending task as currently being deployed.

        @param zone The zone the task is running in.
        @param name The name of the task to flag.
        @param count The max number of task in a zone to flag.
        """
        if name in self.tasks:
            flagged = 0
            for task in self.tasks[name]:
                if not task['machine'] and task['zone'] == zone:
                    flagged += 1
                    task['status'] = self.Deploying
                    if flagged >= count:
                        break

    def get_pending_tasks(self):
        """
        Find the tasks that require machines.

        @param zones Only return required machines for these zones.
        @param tasks Only return required machines for these tasks.
        @return A dictionary describing what tasks in which zones require
            machines. The key is the zone name and the value is another
            dictionary. That dictionary's keys are the task names and the
            values are the number of machines required.
        """
        required = {}
        for name, tasks in self.tasks.iteritems():
            for task in tasks:
                zone = task['zone']
                if zone not in required:
                    required[zone] = {}
                if name not in required[zone]:
                    required[zone][name] = 0
                if not task['machine'] and task['status'] != self.Deploying:
                    required[zone][name] += 1
        return required

    def set_pending_machines(self, zone, name, machines, deploying=False):
        """
        Assign machines to tasks that do not have machines. Each machine will
        be assigned to the given task.

        @param zone The zone to assign to.
        @param name The name of the task to assign machines to.
        @param machines The machines to assign to the task.
        @param deploying True to replace machines that have the deploying flag
        set. Defaults to False.
        """
        for machine in machines:
            for task in self.tasks.get(name, []):
                if (not task['machine'] and task['zone'] == zone and
                        task['status'] == self.Deploying):
                    task['machine'] = machine
                    break

    def get_job_fill(self):
        """
        Calculate the job fill. returns a two-tuple with the job fill and job
        machine fill. These in turn look like this:

        job_fill = {
            job_name: {
                zone: task_count,
            },
        }

        job_fill_machines = {
            job_name: {
                zone: [machines],
            },
        }

        @return The job fill.
        """
        job_fill = {}
        job_fill_machines = {}
        for name, tasks in self.tasks.iteritems():
            job_fill[name] = {}
            job_fill_machines[name] = {}
            for task in tasks:
                if task['zone'] not in job_fill[name]:
                    job_fill[name][task['zone']] = 0
                    job_fill_machines[name][task['zone']] = []
                job_fill[name][task['zone']] += 1
                job_fill_machines[name][task['zone']].append(task['machine'])
        return (job_fill, job_fill_machines)


class ClusterState(object):
    """
    Cluster state. Manages the state of the cluster and generates actions as a
    result of state changes.

    Some static status attributes are defined as part of the class. These are:

    Active -- Machine is operating normally.
    Deploying -- The machine is being deployed.
    Maintenance -- Machine is paused by admin for maintenance.
    Pending -- Waiting for the machine monitor to initialize the machine so
        that tasks can be added to state tracking.
    Unreachable -- Machine is unreachable.
    """

    Active = 'active'
    Deploying = 'deploying'
    Maintenance = 'maint'
    Pending = 'pending'
    Unreachable = 'unreachable'

    def __init__(self, sitter):
        """
        Initialize an empty state object.

        State attributes are structured as follows:

            zones = {
                name: provider,
            }

            machines = [
                {
                    'machine': machine_status,
                    'status': machine_status,
                    'zone': zone,
                },
            ]

        @param sitter The cluster sitter object.
        """
        self.thread = None
        self.running = False
        self.sleep = sitter.stats_poll_interval
        self.sitter = sitter
        self.monitors = []
        self.zones = {}
        self.providers = {}
        self.max_idle_per_zone = -1
        self.machines = []
        self.jobs = {}
        self.job_file = "%s/jobs.json" % sitter.log_location
        self.job_chains = {}
        self.repair_jobs = {}
        self.desired_jobs = JobState()
        self.current_jobs = JobState()
        self.actions = ClusterActionManager()
        self.lock = RLock()

        #TODO: Move this to the sitter. State is not a catch-all.
        self.loggers = []

    @lock
    def _get_machine_item(self, machine):
        """
        Get the list item from self.machines that contains the given machine.
        """
        for item in self.machines:
            if item['machine'] == machine:
                return item
        return None

    @lock
    def _get_master_job(self, job):
        """Get the master job in a job chain."""
        if job.name not in self.jobs:
            return None
        current_job = job
        next_job = job.find_linked_job()
        while next_job:
            current_job = next_job
            next_job = current_job.find_linked_job()
        return current_job

    def get_job(self, name):
        """
        Get a job by name.

        @return The job object or None if it does not exist.
        """
        return self.jobs.get(name)

    def get_zones(self):
        """
        Get the active zones.

        @return A list of active zone names.
        """
        return self.zones.keys()

    def get_providers(self):
        """
        Get the machine providers.

        @return A dictionary of name/provider mappings.
        """
        return dict(self.providers)

    @lock
    def add_provider(self, name, provider):
        """
        Add a machine provider to state tracking. Zones in the provider will
        automatically be added.

        @param name The name of the provider.
        @param provider The provider object.
        """
        if name not in self.providers:
            self.providers[name] = provider
            for zone in provider.get_all_shared_fate_zones():
                self.zones[zone] = name
            return True
        return False

    def get_zone_provider(self, zone):
        """
        Get the provider for a zone.

        @param zone The name of the zone.
        @return The provider for the zone or None if not present.
        """
        return self.providers.get(self.zones.get(zone))

    def has_machine(self, machine):
        """
        Check if the machine is present.

        @param machine The machine to check.
        @return True if present or False.
        """
        return self._get_machine_item(machine) is not None

    def is_machine_mutable(self, machine):
        """
        Check if a machine is allowed to have actions run against it.

        @param machine The machine to check.
        @return True if the machine is mutable or False.
        """
        item = self._get_machine_item(machine)
        disallow_status = [self.Maintenance, self.Pending]
        if not item or item['status'] in disallow_status:
            return False
        return True

    def is_machine_deploying(self, machine):
        """
        Check if a machine is being deployed.

        @param machine The machine to check.
        @return True if the machine is being deployed or False.
        """
        item = self._get_machine_item(machine)
        return item and item['status'] == self.Deploying

    @lock
    def is_machine_idle(self, machine):
        """
        Check if a machine is being deployed.

        @param machine The machine to check.
        @return True if the machine is being deployed or False.
        """
        item = self._get_machine_item(machine)
        return item and self.desired_jobs.is_machine_idle(machine)

    @lock
    def is_machine_monitored(self, machine):
        """
        Check if a machine is being or is queued to be monitored.

        @param machine The machine to check.
        @return True if the machine is monitored or False.
        """
        for monitor, thread in self.monitors:
            if monitor.has_machine(machine):
                return True
        return False

    @lock
    def is_machine_unreachable(self, machine):
        """
        Check if a machine is unreachable.

        @param machine The machine to check.
        @return True if the machine is unreachable or False.
        """
        item = self._get_machine_item(machine)
        return (
            item and not self.is_machine_monitored(machine) and
            not self.get_machine_repair_job(machine))

    @lock
    def get_machines(self, zones=None, status=None, idle=None,
                     unreachable=None):
        """
        Get the machines running in the given zones.

        @param zones The zones to find machines in.
        @param status Only return machines with this status.
        @param idle True to return idle machines, False to return non-idle
            machines. Defaults to both.
        @param unreachable True to return unreachable machine, False to return
            reachable machines. Defaults to both.
        @return A dictionary of zone to machine list mappings.
        """
        if zones:
            zoned_machines = {z: [] for z in zones}
        else:
            zoned_machines = {}

        for item in self.machines:
            zone = item['zone']
            machine = item['machine']
            if item['zone'] not in zoned_machines:
                zoned_machines[item['zone']] = []

            # Check for the requested zone.
            if zones is not None and zone not in zones:
                continue

            # Check for the requested status.
            if status and item['status'] != status:
                continue

            # Check for the requested idle state.
            is_idle = self.is_machine_idle(machine)
            if idle is not None and idle != is_idle:
                continue

            # Check for the requested unreachable state.
            is_unreachable = self.is_machine_unreachable(machine)
            if (unreachable is not None
                    and is_unreachable != unreachable):
                continue

            # Append the machine to the results if all checks passed.
            zoned_machines[zone].append(machine)
        return zoned_machines

    def get_machine_zone(self, machine):
        """
        Get the name of the zone the machine is in.

        @param machine The machine to get the zone for.
        @return The name of the zone the machine is in or None if the machine
            is not present.
        """
        item = self._get_machine_item(machine)
        if not item:
            return None
        return item['zone']

    def get_machine_status(self, machine):
        """
        Get the status of the machine.

        @return The status of the machine or None if the machine is not
            present.
        """
        item = self._get_machine_item(machine)
        if not item:
            return None
        return item['status']

    def get_machine_repair_job(self, machine):
        """
        Get the repair job for the machine.

        @return The repair job for the machine or None if there is no repair
            job.
        """
        return self.repair_jobs.get(machine.hostname)

    @lock
    def add_machine(self, machine, status=None, existing=False):
        """
        Add a machine to state tracking.

        @param machine The machine to add.
        @param zone The zone the machine is in.
        @param status The initial status of the machine. Defaults to Active.
        @param existing True if this is an existing machine with jobs deployed
            to it. Causes the machine to be scanned for task and those tasks
            added to the desired job state.
        @return True on success, False on failure (machine not initialized).
        """
        if not self.has_machine(machine):
            if status is None:
                if existing and not machine.is_initialized():
                    logger.info(
                        "%s not initialized, marked as pending" %
                        machine.hostname)
                    status = self.Pending
                else:
                    status = self.Active

            zone = machine.config.shared_fate_zone
            self.machines.append({
                'machine': machine,
                'status': status,
                'zone': zone,
            })

            if existing and status != self.Pending:
                self.add_machine_tasks(machine)
            return True
        return False

    @lock
    def add_machine_tasks(self, machine):
        """
        Add tasks from a running machine to state tracking.

        @param machine The machine to add tasks from.
        """
        if not machine.is_initialized():
            logger.info(
                "%s not initialized, not adding tasks" %
                machine.hostname)
            return False
        if not machine.has_loaded_data():
            machine.datamanager.reload()
        for task in machine.get_tasks().values():
            name = task['name']
            zone = self.get_machine_zone(machine)
            task_status = JobState.Running
            if not task['running']:
                task_status = JobState.Stopped
            if name not in self.jobs:
                logger.warn((
                    "task '%s' does not have a job object, "
                    "adding anyways") % name)
            self.desired_jobs.add_tasks(
                name, zone, [machine], status=task_status)
            self.current_jobs.add_tasks(
                name, zone, [machine], status=task_status)
        return True

    @lock
    def remove_machine(self, machine):
        """
        Remove a machine from state tracking and monitoring. Tasks associated
        with the machine will also be removed.

        @param machine The machine to remove.
        """
        self.desired_jobs.remove_machine(machine)
        self.current_jobs.remove_machine(machine)
        item = self._get_machine_item(machine)
        if item:
            self.machines.remove(item)

    @lock
    def update_machine(self, machine, status):
        """
        Update the status of a machine.

        @param machine The machine to update.
        @param status The new status of the machine.
        @return True on success or False (machine not found)
        """
        item = self._get_machine_item(machine)
        if item:
            item['status'] = status
            return True
        return False

    @lock
    def monitor_machine(self, machine):
        """
        Add a machine to monitoring.

        @param machine The machine to add to monitoring.
        @return True if the machine was added or False if the machine is
            already being monitored.
        """
        if self.is_machine_monitored(machine):
            return False

        def sort_monitors(monitor):
            return monitor[0].num_monitored_machines()

        # Always add to the monitor with the least amount of machines.
        self.monitors.sort(key=sort_monitors)
        return self.monitors[0][0].add_machine(machine)

    @lock
    def unmonitor_machine(self, machine):
        """
        Remove a machine from monitoring.

        @param machine The machine to remove from monitoring.
        @return True if the machine was removed or False if the machine was not
            being monitored.
        """
        for monitor, thread in self.monitors:
            if monitor.remove_machine(machine):
                return True
        return False

    @lock
    def repair_machine(self, machine):
        """
        Redeploy a machinesitter to a machine.

        @param machine The machine to repair.
        @return True if a repair job was generated or False if one already
            exists.
        """
        #TODO: Handle machine repairs in machine object.
        if machine.hostname in self.repair_jobs:
            return False
        self.update_machine(machine, self.Deploying)

        zone = self.get_machine_zone(machine)
        job = ProductionJob(
            self.sitter,
            '', {'name': 'Machine Redeployer'}, {
                zone: {
                    'mem': machine.config.mem,
                    'cpu': machine.config.cpus,
                },
            }, None)

        self.repair_jobs[machine.hostname] = job
        self.actions.add(
            RedeployMachineAction(self.sitter, zone, machine, job))
        return True

    @lock
    def detach_machine(self, machine):
        """
        Undeploy all jobs from a machine. Jobs removed in this manner will be
        assigned to another idle job. Used to keep jobs around when undeploying
        a dead machine.

        @param machine The machine to detach.
        """
        zone = self.get_machine_zone(machine)
        tasks = self.desired_jobs.get_machine_tasks(machine)
        job_chains = {}
        for task in tasks:
            if task in self.job_chains:
                job_chains[task] = self.job_chains[task]
            self.desired_jobs.remove_tasks(task, [machine])

        for master in job_chains.keys():
            self.desired_jobs.add_tasks(master, zone, [], 1)

    @lock
    def add_job(self, job, redeploy=False):
        """
        Add a job to the cluster. Master jobs are immediately allocated to
        machines. Child jobs are attached to their associated master job and
        deployed simultenously as a job "chain". Existing jobs will be
        redeployed if requested. If the name of this job matches existing tasks
        that do not have jobs then those tasks will be assigned to this job.

        @param job The job to add to the cluster. The job is configured with
            the required zones and number of machines.
        @param redeploy Whether or not to redeploy existing jobs.
        @return True if the job was added or False if it was not.
        """
        self.jobs[job.name] = job
        zoned_existing_machines = self.desired_jobs.get_task_machines(job.name)
        zoned_idle_machines = self.get_machines(idle=True)
        job_zones = job.get_shared_fate_zones()
        all_zones = set(job_zones) | set(zoned_existing_machines.keys())

        for zone in all_zones:
            existing_machines = zoned_existing_machines.get(zone, [])

            if zone not in job_zones:
                # Remove the zone if it's not in the job.
                self.desired_jobs.remove_tasks(
                    job.name, existing_machines)
            elif job.linked_job:
                # Redeploy child tasks to their current machines.
                if redeploy:
                    self.desired_jobs.update_tasks(
                        job.name, JobState.Deploying, existing_machines)
                    for machine in existing_machines:
                        self.actions.add(
                            AddTaskAction(self.sitter, zone, machine, job.name))
            else:
                # Redeploy master job across the cluster.
                idle_machines = zoned_idle_machines.get(zone, [])
                num_required = job.get_num_required_machines_in_zone(zone)

                redeploy_machines = existing_machines[:num_required]
                undeploy_machines = existing_machines[num_required:]
                num_idle = num_required - len(redeploy_machines)
                deploy_machines = idle_machines[:num_idle]
                num_create = (
                    num_required - len(redeploy_machines) -
                    len(deploy_machines))

                self.desired_jobs.add_tasks(
                    job.name, zone, deploy_machines, num_create)
                self.desired_jobs.remove_tasks(
                    job.name, undeploy_machines)

                if redeploy:
                    self.desired_jobs.update_tasks(
                        job.name, JobState.Deploying, redeploy_machines)
                    for machine in redeploy_machines:
                        self.actions.add(
                            AddTaskAction(self.sitter, zone, machine, job.name))
        return True

    @lock
    def update_job(self, job, version=None):
        """
        Update the job. The job is redeployed to existing machines.

        @param job The job or name of the job to update.
        @param version The version to update to. Optional.
        @return True if successful or False if the job does not exist.
        """
        name = str(job)
        job = self.get_job(name)

        if not job:
            return False
        job.do_update_deployment(self, version)
        return True

    @lock
    def remove_job(self, job):
        """
        Remove a job from the cluster. Child jobs will be removed as well.

        @param job The job or name of the job to remove.
        @return A list of job names that were removed.
        """
        tasks = []
        if isinstance(job, basestring):
            tasks.append(job)
            job = self.jobs.get(job)
        else:
            tasks.append(job.name)
            children = job.find_dependent_jobs()
            for child in children:
                tasks.append(child.name)

        removed = []
        for task in tasks:
            removed.extend(self.desired_jobs.remove_tasks(task))
        return removed

    @lock
    def persist_jobs(self):
        """
        Persist jobs to the jobs file.
        """
        jobs = [j for j in self.jobs.values() if j.persistent]
        with open(self.job_file, 'w') as fd:
            fd.write(simplejson.dumps([j.to_dict() for j in jobs]))

    @lock
    def start_task(self, machine, task):
        """
        Start a task on a machine.

        @param machine The machine on which the task is running.
        @param task The name of the task to start.
        @return True if successful or False (job not deployed)
        """
        if task not in self.jobs:
            logger.warn("job '%s' not deployed, not starting" % task)
            return False

        self.desired_jobs.update_tasks(task, JobState.Running, [machine])
        return True

    @lock
    def stop_task(self, machine, task):
        """
        Stop a task on a machine.

        @param machine The machine on which the task is running.
        @param task The name of the task to stop.
        @return True if successful or False (job not deployed)
        """
        if task not in self.jobs:
            logger.warn("job '%s' not deployed, not stopping" % task)
            return False

        self.desired_jobs.update_tasks(task, JobState.Running, [machine])
        return True

    @lock
    def restart_task(self, machine, task):
        """
        Restart a task on a machine.

        @param machine The machine on which the task is running.
        @param task The name of the task to restart.
        @return True if successful or False (job not deployed)
        """
        if task not in self.jobs:
            logger.warn("job '%s' not deployed, not restarting" % task)
            return False

        status = self.desired_jobs.get_task_status(task, machine)
        if status is None:
            logger.warn(
                "job '%s' not deployed to '%s', not restarting" %
                (task, machine.hostname))
            return False

        if status != JobState.Running:
            self.desired_jobs.update_tasks(task, JobState.Running, [machine])
        else:
            if not self.is_machine_mutable(machine):
                logger.info(
                    "machine '%s' in maintenance mode, not restarting" %
                    machine.hostname)
                return False

            zone = self.get_machine_zone(machine)
            self.actions.add(
                RestartTaskAction(self.sitter, zone, machine, task))
            logger.info(
                "restart action queued for task '%s' on machine '%s'" %
                (task, machine.hostname))
        return True

    def calculate_ready_machines(self):
        """
        Convert ready pending machines to active.
        """
        for machine in self.machines:
            if (machine['status'] == self.Pending and
                    machine['machine'].is_initialized()):
                if self.add_machine_tasks(machine['machine']):
                    machine['status'] = self.Active

    def calculate_current_state(self):
        """
        Update the current state of jobs on the cluster.
        """
        #TODO: Check providers for new zones.
        zoned_machines = self.get_machines()
        job_machines = self.current_jobs.get_task_machines()

        # Remove missing machines.
        for zone, machines in job_machines.iteritems():
            for machine in machines:
                if (zone not in zoned_machines or
                        machine not in zoned_machines[zone]):
                    self.current_jobs.remove_machine(machine)

        # Update tasks.
        for zone, machines in zoned_machines.iteritems():
            for machine in machines:
                if not machine.is_initialized():
                    logger.info(
                        "machine '%s' not initialized, skipping state update" %
                        machine.hostname)
                    continue

                expected_tasks = self.current_jobs.get_machine_tasks(machine)
                actual_tasks = {}
                for actual_task in machine.get_tasks().values():
                    if actual_task['running']:
                        actual_tasks[actual_task['name']] = JobState.Running
                    else:
                        actual_tasks[actual_task['name']] = JobState.Stopped

                add_tasks = set(actual_tasks.keys()) - set(expected_tasks)
                remove_tasks = set(expected_tasks) - set(actual_tasks.keys())
                update_tasks = set(expected_tasks) & set(actual_tasks.keys())

                for task in add_tasks:
                    self.current_jobs.add_tasks(
                        task, zone, [machine], status=actual_tasks[task])
                for task in remove_tasks:
                    self.current_jobs.remove_tasks(task, [machine])
                for task in update_tasks:
                    self.current_jobs.update_tasks(
                        task, actual_tasks[task], [machine])

    def calculate_job_chains(self):
        """
        Generate job chains for use in job deployment.
        """
        # Build job chains.
        job_chains = {}
        for name, job in self.jobs.iteritems():
            master = self._get_master_job(job)
            if not master:
                continue
            if master.name not in job_chains:
                job_chains[master.name] = []
            if master.name != job.name:
                job_chains[master.name].append(job.name)
        self.job_chains = job_chains

    def calculate_job_deployment(self):
        """
        Generate the actions necessary to convert the current job state to the
        desired job state.
        """
        # Assign chain tasks to machines.
        for master, children in self.job_chains.iteritems():
            # Find machines where the master is running.
            zoned_master_machines = self.desired_jobs.get_task_machines(master)
            for zone, master_machines in zoned_master_machines.iteritems():
                for child in children:
                    # Assign child tasks.
                    machines = []
                    for machine in master_machines:
                        tasks = self.desired_jobs.get_machine_tasks(machine)
                        if child not in tasks:
                            machines.append(machine)
                    self.desired_jobs.add_tasks(child, zone, machines)

        # Assign idle machines to pending tasks.
        pending_tasks = self.desired_jobs.get_pending_tasks()
        for zone, tasks in pending_tasks.iteritems():
            for master, required in tasks.iteritems():
                if required > 0:
                    master_job = self.get_job(master)

                    # Find some idle machines.
                    idle_machines = self.get_machines(
                        zones=[zone], idle=True)[zone]
                    idle_machines = idle_machines[:required]
                    required = required - len(idle_machines)

                    # Assign idle machines to task chain.
                    self.desired_jobs.set_pending_machines(
                        zone, master, idle_machines)

                    # Deploy master job to new machines.
                    self.desired_jobs.set_pending_deploying(
                        zone, master, required)
                    for n in range(required):
                        self.actions.add(
                            DeployMachineAction(self.sitter, zone, master_job))

        # Calculate changes to existing tasks.
        def filter_valid_tasks(tasks):
            return set([t for t in tasks if t[2] in self.jobs])
        desired_tasks = filter_valid_tasks(self.desired_jobs.flatten())
        current_tasks = filter_valid_tasks(self.current_jobs.flatten())

        add_tasks = desired_tasks - current_tasks
        remove_tasks = current_tasks - desired_tasks
        check_tasks = desired_tasks & current_tasks

        # Add actions for existing tasks.
        for task in add_tasks:
            if not self.desired_jobs.is_task_deploying(task[2], task[1]):
                self.desired_jobs.update_tasks(
                    task[2], JobState.Deploying, [task[1]])
                self.actions.add(AddTaskAction(self.sitter, *task))

        for task in remove_tasks:
            #TODO: Remove status check when undeploying job code works.
            if (self.current_jobs.get_task_status(task[2], task[1]) ==
                    JobState.Stopped):
                continue
            if self.is_machine_mutable(task[1]):
                self.actions.add(RemoveTaskAction(self.sitter, *task))

        for task in check_tasks:
            if (not self.desired_jobs.is_task_deploying(task[2], task[1])
                    and self.is_machine_mutable(task[1])):
                desired_status = self.desired_jobs.get_task_status(
                    task[2], task[1])
                current_status = self.current_jobs.get_task_status(
                    task[2], task[1])
                if desired_status != current_status:
                    if desired_status == JobState.Running:
                        self.actions.add(StartTaskAction(self.sitter, *task))
                    else:
                        self.actions.add(StopTaskAction(self.sitter, *task))

    def calculate_idle_cleanup(self):
        """
        Calculate the idle machines and decomission the ones we don't need.
        """
        if self.max_idle_per_zone >= 0:
            decomission = {}
            zoned_idle_machines = self.get_machines(idle=True)
            for zone, idle_machines in zoned_idle_machines.iteritems():
                if len(idle_machines) > self.max_idle_per_zone:
                    count = len(idle_machines) - self.max_idle_per_zone
                    decomission[zone] = idle_machines[-count:]

            for zone, machines in decomission.iteritems():
                for machine in machines:
                    logger.info(
                        "decomissioning idle machine '%s'" %
                        machine.hostname)
                    self.actions.add(
                        DecomissionMachineAction(self.sitter, zone, machine))

    def calculate_job_cleanup(self):
        """
        Remove jobs that are no longer deployed to machines. If the job is a
        child job then we remove it if its master is no longer deployed to any
        machines. Also remove finished repair jobs.
        """
        for job in self.jobs.values():
            master = self._get_master_job(job)
            if (master and not self.desired_jobs.has_task(master.name) and
                    not self.current_jobs.has_task(master.name)):
                del self.jobs[job.name]
        self.persist_jobs()

        for name, job in dict(self.repair_jobs).iteritems():
            remove = True
            for fillers in job.fillers.values():
                for filler in fillers:
                    # If there are fillers still running then leave it in.
                    if filler.state.get_state() != 8:
                        remove = False
                        break
            if remove:
                del self.repair_jobs[name]

    def calculate_unreachable_machines(self):
        """
        Redeploy sitters to unreachable machines.
        """
        zoned_machines = self.get_machines()
        for zone, machines in zoned_machines.iteritems():
            for machine in machines:
                if (not self.is_machine_monitored(machine) and
                        not self.get_machine_repair_job(machine)):
                    self.repair_machine(machine)

    @lock
    def calculate(self):
        """
        Handle all state calculations.

        @return Actions generated by the state calculations.
        """
        self.calculate_ready_machines()
        self.calculate_current_state()
        self.calculate_job_chains()
        self.calculate_job_deployment()
        self.calculate_job_cleanup()
        self.calculate_idle_cleanup()
        self.calculate_unreachable_machines()

    def process(self):
        """
        Process actions asynchronously.
        """
        self.actions.process()

    def run(self):
        """
        Run a full calculate/process cycle.
        """
        try:
            logger.info("start calculation cycle")
            self.calculate()
            self.process()
        except:
            import traceback
            logger.error("failure during state cycle")
            logger.error(traceback.format_exc())

    def start(self):
        """
        Run the state calculator until stopped.
        """
        def run_loop():
            while self.running:
                self.run()
                time.sleep(self.sleep)

        if self.thread is None or not self.thread.is_alive():
            self.thread = Thread(target=run_loop, name="Calculator")
        self.running = True
        self.thread.start()

    def stop(self, timeout=None):
        """
        Stop the state calculator.
        """
        self.running = False
        self.thread.join()



import curses
import os
import subprocess
import sys

from datetime import datetime
from menu import MenuFactory, MenuOption, MenuChanger, Table
from sittercommon.machinedata import MachineData
from sittercommon.utils import strip_html

MACHINE_DATA = None
SCREEN = None


class ManagementScreen(object):

    def __init__(self):
        self.factory = None
        self.scr = None
        self.ypos = 0
        self.current_loc = "mainmenu"
        self.aux = None

    def add_line(self, msg):
        self.scr.addstr(self.ypos, 0, msg)
        self.ypos += 1

    def remove_line(self):
        self.scr.clrtoeol()
        y, x = self.scr.getyx()
        self.scr.move(y - 1, x)
        self.ypos -= 1

    def refresh(self):
        self.scr.clear()
        self.scr.refresh()
        self.ypos = 0

    def change_menu(self, newmenu, aux=None):
        self.current_loc = newmenu
        if aux:
            self.aux = aux

    def header(self):
        pass

    def basic_tasks(self):
        pass

    def mainmenu(self):
        pass

    def reload_data(self):
        pass

    def run(self):
        self.scr = curses.initscr()
        self.refresh()

        while True:
            self.refresh()
            self.header()
            self.factory = MenuFactory(self.scr,
                                       self.add_line,
                                       self.remove_line)

            self.add_line("-" * self.scr.getmaxyx()[1])
            self.basic_tasks()
            self.add_line("-" * self.scr.getmaxyx()[1])
            if hasattr(self, self.current_loc):
                getattr(self, self.current_loc)()
            else:
                globals()[self.current_loc]()


class MachineManagementScreen(ManagementScreen):
    def __init__(self, machine_data, *args, **kwargs):
        super(MachineManagementScreen, self).__init__(*args, **kwargs)
        self.machine_data = machine_data

    def header(self):
        self.add_line("#" * self.scr.getmaxyx()[1])
        self.add_line(
            "MachineSitter at %s - Curses UI %s" % (
                self.machine_data.url, datetime.now()))
        self.add_line("#" * self.scr.getmaxyx()[1])

    def reload_data(self):
        self.machine_data.reload()

    def mainmenu(self):
        menu = self.factory.new_menu("Main Menu")
        menu.add_option_vals("Main Menu", hotkey="*",
                             action=lambda: self.change_menu('mainmenu'))
        menu.add_option_vals("Add a new task",
                             action=lambda: self.change_menu('addtask'))

        menu.add_option_vals(
            "Show task definitions",
            action=lambda: self.change_menu(
                'show_task_definitions'))

        menu.add_option_vals(
            "Show machine sitter logs",
            action=lambda: self.change_menu(
                'show_machinesitter_logs'))

        menu.add_option_vals("Stop machine sitter",
                             action=stop_sitter)

        menu.render()

    def basic_tasks(self):
        self.reload_data()
        running = []
        not_running = []
        not_running_lines = []
        for name, task in self.machine_data.tasks.items():
            if task['running']:
                running.append((name, task))
            else:
                not_running.append((name, task))

        hotkey = 1

        table = Table(self.scr.getmaxyx()[1],
                      ["Running Task Name", "Restarts",
                       "Runtime", "Stdout", "Stderr"])

        for name, task in running:
            runtime = '?'
            stdout_kb = -1.0
            stderr_kb = -1.0
            if task.get('process_start_time'):
                runtime = str(
                    datetime.now() - datetime.strptime(
                        task['process_start_time'],
                        '%Y-%m-%d %H:%M:%S.%f')
                )

                # strip useconds
                runtime = runtime[:runtime.find('.')]
                try:
                    stdout = self.machine_data.get_logfile(task)['location']
                    stderr = self.machine_data.get_logfile(
                        task, True)['location']

                    stdout_kb = os.stat(stdout).st_size / 1024
                    stderr_kb = os.stat(stderr).st_size / 1024
                except:
                   pass

            table.add_row([task['name'],
                           task.get('num_task_starts', '?'),
                           runtime,
                           "%.0f kB" % stdout_kb,
                           "%.0f kB" % stderr_kb])

            option = MenuOption(
                task['name'],
                action=MenuChanger(self.change_menu, "show_task",
                                   (name, task)),
                hotkey=str(hotkey),
                hidden=True)

            self.factory.add_default_option(option)
            hotkey += 1

        for name, task in not_running:
            line = task['name']
            not_running_lines.append(line)
            option = MenuOption(
                task['name'],
                action=MenuChanger(self.change_menu, "show_task",
                                   (name, task)),
                hotkey=str(hotkey),
                hidden=True)

            self.factory.add_default_option(option)
            hotkey += 1

        table.render(self)

        self.add_line("Stopped Tasks:")
        for num, l in enumerate(not_running_lines):
            self.add_line("%s. %s" % ((len(running) + num + 1), l))

    def show_task(self):
        name, task = self.aux
        self.reload_data()
        task = self.machine_data.tasks[name]

        menu = self.factory.new_menu(
            "%s (%s) (%s)" % (
                name,
                task['command'],
                strip_html(task.get('monitoring', 'Not Running'))
            ))

        menu.add_option_vals(
            "Main Menu",
            action=lambda: self.change_menu(
                'mainmenu'), hotkey="*")

        if not task['running']:
            menu.add_option_vals("Start Task",
                                 action=lambda: self.start_task(task))
            menu.add_option_vals("Remove Task",
                                 action=lambda: self.remove_task(task))
        else:
            menu.add_option_vals("Stop Task",
                                 action=lambda: self.stop_task(task))
            menu.add_option_vals("Restart Task",
                                 action=lambda: self.restart_task(task))

            menu.add_option_vals("Show stdout/stderr",
                                 action=lambda: self.show_log(task))

        menu.add_option_vals(
            "Show historic task log files",
            action=lambda: self.change_menu('show_task_logs', task))

        menu.render()

    def show_log(self, task):
        tail_file([
            self.machine_data.get_logfile(task)['location'],
            self.machine_data.get_logfile(task, True)['location']])

    def show_machinesitter_logs(self):
        logs = self.machine_data.get_sitter_logs()
        self.show_logs(logs, "Machine Sitter Logs")

    def show_task_logs(self):
        task = self.aux
        logs = self.machine_data.get_task_logs(task)
        self.show_logs(logs, "%s Logs" % task['name'])

    def show_logs(self, logs, title):
        menu = self.factory.new_menu(title)
        menu.add_option_vals(
            "Main Menu",
            action=lambda: self.change_menu('mainmenu'),
            hotkey="*")

        lognames = logs.keys()
        lognames.sort()

        for logname in lognames:
            logfile = logs[logname]['location']
            menu.add_option_vals("%s (%s)" % (logname, logfile),
                                 action=MenuChanger(tail_file, [logfile]))

        menu.render()

    def start_task(self, task):
        self.machine_data.start_task(task)

    def stop_task(self, task):
        self.machine_data.stop_task(task)

    def restart_task(self, task):
        self.machine_data.restart_task(task)

    def remove_task(self, task):
        self.machine_data.remove_task(task)
        self.current_loc = "mainmenu"


def stop_sitter():
    os.kill(int(MACHINE_DATA.metadata['machinesitter_pid']), 15)
    sys.exit(0)


def tail_file(filenames):
    curses.endwin()
    os.system("clear")
    subprocess.call(["echo", ' '.join(filenames)])
    try:
        args = ["tail", "-n", "100", "-f"]
        args.extend(filenames)
        subprocess.call(args)
    except:
        pass


def main():
    global MACHINE_DATA, SCREEN
    if not MACHINE_DATA:
        MACHINE_DATA = MachineData("localhost", 40000)

    SCREEN = MachineManagementScreen(MACHINE_DATA)

    try:

        if not MACHINE_DATA.url:
            print "Couldn't find a running machine sitter!"
            return

        SCREEN.run()
    except:
        curses.endwin()
        import traceback
        traceback.print_exc()

if __name__ == '__main__':
    main()



import json
import os
import signal
import simplejson
import socket
import threading
import time

import sittercommon.http_monitor as http_monitor
import sittercommon.logmanager as logmanager
import machinestats
import taskmanager


class MachineManager(object):

    def __init__(self, task_definition_file, log_location,
                 machine_sitter_starting_port=40000,
                 task_sitter_starting_port=50000,
                 launch_location="",
                 daemon=False):
        self.tasks = {}
        self.launch_location = launch_location
        self.task_definition_file = task_definition_file
        self.daemon = daemon
        self.thread = None
        self.should_stop = False

        self.log_location = log_location
        self.start_count = 0
        self.command = "machinemanager"
        self.logmanager = logmanager.LogManager(stdout_location=log_location,
                                                stderr_location=log_location)
        self.logmanager.set_harness(self)

        self.parent_pid = os.getpid()
        self.stats = machinestats.MachineStats(self)

        self.machine_sitter_starting_port = machine_sitter_starting_port
        self.task_sitter_starting_port = task_sitter_starting_port
        self.orig_machine_port = self.machine_sitter_starting_port
        self.orig_task_port = self.task_sitter_starting_port

        self.http_monitor = http_monitor.HTTPMonitor(self.stats,
                                                     self,
                                                     self.next_port(True))

        self.http_monitor.add_handler('/start_task', self.remote_start_task)
        self.http_monitor.add_handler('/stop_task', self.remote_stop_task)
        self.http_monitor.add_handler('/remove_task', self.remote_remove_task)
        self.http_monitor.add_handler('/add_task', self.remote_add_task)
        self.http_monitor.add_handler('/restart_task',
                                      self.remote_restart_task)
        self.http_monitor.add_handler('/write_config',
                                      self.remote_write_config)
        self.http_monitor.add_handler('/load_config',
                                      self.remote_load_config)

        print "Adding signals"
        signal.signal(signal.SIGTERM, self.exit_now)
        signal.signal(signal.SIGINT, self.exit_now)

    def exit_now(self, *_):
        print "Caught Control-C, killing children and exiting"
        for task in self.tasks.values():
            if task.is_running():
                task.stop()

        os._exit(0)

    def remote_add_task(self, args):
        definition = {}
        for opt in taskmanager.TaskManager.required_fields:
            definition[opt] = args.get(opt)
            if not definition[opt]:
                return "Required argument %s missing" % opt
        for opt in taskmanager.TaskManager.optional_fields:
            if opt in args:
                definition[opt] = args[opt]

        for k, v in definition.items():
            if v.lower() == 'true':
                definition[k] = True
            if v.lower() == 'false':
                definition[k] = False

        task = self.add_new_task(definition)
        task.initialize()

        config = self.write_out_task_definitions()

        return "OK | %s" % json.dumps(config)

    def remote_remove_task(self, args):
        if not 'task_name' in args:
            return "Error, no task name provided"

        if not args['task_name'] in self.tasks:
            return "Error, unknown task"

        task = self.tasks[args['task_name']]
        if task.was_started:
            return "You must stop a task before you remove it"

        del self.tasks[task.name]
        return "Removed"

    def remote_stop_task(self, args):
        if not 'task_name' in args:
            return "Error, no task name provided"

        if not args['task_name'] in self.tasks:
            return "Error, unknown task"

        task = self.tasks[args['task_name']]
        if not task.was_started:
            return "Already stopped"

        task.stop()
        self.collect_old_task_logs(task)
        return "Stopped"

    def remote_restart_task(self, args):
        if not 'task_name' in args:
            return "Error, no task name provided"

        if not args['task_name'] in self.tasks:
            return "Error, unknown task"

        task = self.tasks[args['task_name']]
        if task.was_started:
            task.stop()

        self.collect_old_task_logs(task)
        task.set_port(self.next_port())
        task.start()

        return "Restarted"

    def remote_start_task(self, args):
        if not 'task_name' in args:
            return "Error"

        if not args['task_name'] in self.tasks:
            return "Error"

        task = self.tasks[args['task_name']]
        if not task.is_running():
            task.set_port(self.next_port())
            task.start()
            return "%s started" % args['task_name']
        else:
            return "Already running"

    def remote_write_config(self, args):
        if not 'file' in args:
            return "Error, no file provided"

        self.task_definition_file = args['file']
        return "Wrote: %s" % self.write_out_task_definitions()

    def remote_load_config(self, args):
        if not 'file' in args:
            return "Error, no file provided"

        self.task_definition_file = args['file']

        data = {}
        try:
            taskfile = open(self.task_definition_file)
            data = simplejson.loads(taskfile.read())
            taskfile.close()
        except:
            import traceback
            return traceback.format_exc()

        for task in data['task_definitions']:
            self.add_new_task(task)

        return "Added tasks: %s" % data

    def write_out_task_definitions(self):
        config = {'log_location': self.log_location}
        task_definitions = []
        for task in self.tasks:
            task_definitions.append(self.tasks[task].to_dict())

        config['task_definitions'] = task_definitions

        if self.task_definition_file:
            fileh = open(self.task_definition_file, 'w')
            fileh.write(json.dumps(config))
            fileh.close()

        return config

    def next_port(self, use_machine_port=False):
        works = None
        starting_port = self.task_sitter_starting_port
        orig_port = self.orig_task_port

        if use_machine_port:
            starting_port = self.machine_sitter_starting_port
            orig_port = self.orig_machine_port

        while not works:
            try:
                sok = socket.socket(socket.AF_INET)
                sok.bind(("0.0.0.0", starting_port))
                sok.close()
                works = starting_port
                starting_port += 1
            except:
                starting_port += 1
                if starting_port > orig_port + 10000:
                    starting_port = orig_port

        if use_machine_port:
            self.machine_sitter_starting_port = starting_port
        else:
            self.task_sitter_starting_port = starting_port

        return works

    def add_new_task(self, task_definition):
        task = taskmanager.TaskManager(task_definition,
                                       self.log_location,
                                       self.launch_location)
        if task.name in self.tasks:
            # If we're adding a task that already exists, simply
            # update the existing object instead of overwriting it.
            task = self.tasks[task.name]
            task.reload_from_definition(task_definition)

        task.set_port(self.next_port())

        self.tasks[task.name] = task
        return task

    def start_task(self, task_name):
        self.tasks[task_name].start()

    def start(self):
        self.thread = threading.Thread(target=self._run)
        self.thread.start()

    def stop(self):
        self.should_stop = True

    def collect_old_task_logs(self, task):
        for filename, fileloc in task.get_old_logfilenames().items():
            self.logmanager.add_logfile("Terminated %s (%s @ %s) %s" % (
                    task.name,
                    task.command,
                    task.get_last_pid(),
                    filename
                    ), fileloc)

    def task_finished(self, task):
        print "Task %s finished" % task.name
        logs = task.stdall()
        print "Task Stdout:\n %s" % logs[0]
        print "Task Stderr:\n %s" % logs[1]
        self.collect_old_task_logs(task)

    def restart_task(self, task):
        task.stop()
        task.set_port(self.next_port())
        task.start()

    def _run(self):
        self.http_monitor.start()
        print "Machine Sitter Monitor started at " + \
            "http://localhost:%s" % self.http_monitor.port

        stdout_loc = self.logmanager._calculate_filename(
            self.logmanager.stdout_location)
        stderr_loc = self.logmanager._calculate_filename(
            self.logmanager.stderr_location, True)
        self.logmanager.add_logfile("machinemanager-stdout", stdout_loc)
        self.logmanager.add_logfile("machinemanager-stderr", stderr_loc)

        if self.daemon:
            self.logmanager.setup_all()

        for task in self.tasks.values():
            print "Initializing %s" % task.name
            task.initialize()
            self.logmanager.add_logfile(
                "%s-stdout" % task.name, task.sitter_stdout)

            self.logmanager.add_logfile(
                "%s-stderr" % task.name, task.sitter_stderr)

        while True:
            for task in self.tasks.values():
                if task.was_started and not task.is_running():
                    print "%s: started: %s, running: %s" % (
                        task, task.was_started, task.is_running())
                    self.task_finished(task)
                    if not task.allow_exit:
                        self.restart_task(task)
                    else:
                        task.was_started = False

            time.sleep(1)



"""
Define a class which represents a constraint on the sub task
"""
import datetime


class Constraint(object):
    """A Constraint on a task."""
    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.kill_on_violation = True

    def check_violation(self, child_proc):
        """
        Look for a violation of a constraint

        Args: child_proc - a Process object

        Returns: True if in violation
        """
        print "check generic violation %s" % child_proc.pid
        return 0


class LivingConstraint(Constraint):
    """
    Ensure the child process is alive.
    """
    def __init__(self):
        super(LivingConstraint, self).__init__("Child is alive? Constraints",
                                               True)
        self.kill_on_violation = False

    def check_violation(self, child_proc):
        """
        Ensure the child process is alive.
        """
        return child_proc.is_alive() != self.value

    def __str__(self):
        return "LivingConstraint"


class TimeConstraint(Constraint):
    """
    Only allow a child process to live so long
    """
    def __init__(self, time_limit):
        super(TimeConstraint, self).__init__("Time Limit Constraint",
                                               datetime.timedelta(
                seconds=time_limit))

    def check_violation(self, child_proc):
        """
        Check how long a child has been alive.
        """
        now = datetime.datetime.now()
        return now - child_proc.start_time > self.value

    def __str__(self):
        return "TimeConstraint (%ss)" % self.value.seconds


class CPUConstraint(Constraint):
    """
    Ensure we satisfy a CPU based constraint.
    """
    def __init__(self, cpu_limit):
        super(CPUConstraint, self).__init__("CPU Based Constraint",
                                             cpu_limit)

    def check_violation(self, child_proc):
        """
        Calculate child CPU usage and see if it violates the constraint.
        """
        child_proc.update_usage(deep=True)

        if child_proc.cpu_usage > float(self.value):
            print "CPU Limit Exceeded"
            return True

        return False

    def __str__(self):
        return "CPU Constraint (%s)" % self.value


class MemoryConstraint(Constraint):
    """
    Ensure we satisfy a memory based constraint
    """
    def __init__(self, mem_limit):
        # convert MB to bytes
        mem_limit = int(mem_limit) * 1024 * 1024
        super(MemoryConstraint, self).__init__("Memory Based Constraint",
                                             mem_limit)

    def check_violation(self, child_proc):
        """ Check for using too much Memory"""
        child_proc.update_usage(deep=True)

        if child_proc.mem_usage[1] > self.value:
            print "Memory Limit Exceeded"
            return True
        return False

    def __str__(self):
        return "Memory Constraint (%s MB)" % (self.value / 1024 / 1024)



class Table(object):
    def __init__(self, max_width, headers):
        self.headers = headers
        self.rows = []
        self.column_widths = []
        self.max_width = max_width

    def add_row(self, items):
        if len(items) != len(self.headers):
            raise Exception("Invalid number of items in this row!")
        self.rows.append(items)

    def render_cell(self, columnnum, rownum, column):
        col_len = len(str(column))
        padding = int((self.column_widths[columnnum] - col_len))
        if rownum > 9 and columnnum == 0:
            padding -= 1

        return "%s%s" % (column, " " * padding)

    def __str__(self):
        self.column_widths = [0] * len(self.headers)
        self.rows = [self.headers] + self.rows
        for column in range(len(self.headers)):
            for row in self.rows:
                if len(str(row[column])) > self.column_widths[column]:
                    self.column_widths[column] = len(str(row[column]))

        spacer = '|  '
        # Don't forget to account for the bars!
        total_width = sum(self.column_widths) + (
            len(spacer) * len(self.column_widths))

        if total_width < self.max_width:
            for i in range(len(self.column_widths)):
                self.column_widths[i] += (
                    (self.max_width - total_width) / len(self.column_widths))

        lines = []
        for rownum, row in enumerate(self.rows):
            line = spacer.join(
                [self.render_cell(
                    i, rownum, cell) for i, cell in enumerate(row)])
            lines.append(line)
        return '\n'.join(lines)

    def render(self, console):
        show_table = str(self).split('\n')
        console.add_line("   %s" % show_table[0])

        for num, l in enumerate(show_table[1:]):
            console.add_line("%s. %s" % ((num + 1), l))


class MenuOption(object):
    def __init__(self, value, action, hotkey=None, hidden=False):
        self.value = value
        self.action = action
        self.hotkey = hotkey
        self.hidden = hidden

    def __str__(self):
        return "%s. %s" % (self.hotkey, self.value)


class MenuChanger(object):
    def __init__(self, changer, *args):
        self.changer = changer
        self.args = args

    def __call__(self):
        self.changer(*self.args)


class Menu(object):
    def __init__(self, title, screen, writer, remover):
        self.title = title
        self.options = []
        self.page = 0
        self.items_per_page = screen.getmaxyx()[0] - 11
        self.screen = screen
        self.writer = writer
        self.remover = remover
        self.written_lines = 0

    def add_option(self, option):
        self.options.append(option)

    def add_option_vals(self, value, action, hotkey=None, hidden=False):
        self.options.append(MenuOption(value, action, hotkey, hidden))

    def next_page(self):
        self.page += 1
        self.clear()
        self.render()

    def prev_page(self):
        self.page -= 1
        self.clear()
        self.render()

    def clear(self):
        for _ in range(self.written_lines):
            self.remover()

        self.written_lines = 0

    def write(self, msg):
        self.writer(msg)
        self.written_lines += 1

    def render(self):
        hotkeys = ['q', 'w', 'e', 'r', 't', 'y', 'a', 's', 'd', 'f',
                   'g', 'z', 'x', 'c', 'v', 'b', 'y', 'u', 'i', 'o',
                   'h', 'j', 'k', 'l']
        hotkeys *= 30
        hotkeys.reverse()
        option_list = [o for o in self.options]

        # Count how many are hidden so we know how many
        # we actually can display
        hidden_count = 0
        for option in option_list:
            if option.hidden:
                hidden_count += 1

            if not option.hotkey:
                option.hotkey = hotkeys.pop()
        shown_options = (len(self.options) - hidden_count)
        max_per_page = (self.items_per_page - hidden_count) or 1
        pages = int(shown_options / max_per_page) + 1

        # Add a page count to the title of necessary
        if pages == 1:
            self.write(" " * 20 + self.title)
        else:
            self.write(" " * 20 + self.title + " (%s/%s)" % (
                self.page + 1, pages))

        index = 0
        for option in option_list:
            if not option.hidden:
                if (index >= (self.page * max_per_page) and
                        index < ((self.page + 1) * max_per_page)):

                    self.write(str(option))

                index += 1

        # Add the next/prev buttons
        if shown_options > max_per_page:
            if self.page != (pages - 1):
                opt = MenuOption('Next Page',
                                 self.next_page, 'n')
                option_list.append(opt)
                self.write(str(opt))

            if self.page != 0:
                opt = MenuOption('Prev. Page', self.prev_page, 'p')
                option_list.append(opt)
                self.write(str(opt))

        self.write("Your Choice: ")
        self.screen.refresh()
        char = self.screen.getstr()

        if not char:
            return

        for option in option_list:
            if str(char) == str(option.hotkey):
                return option.action()

        # We didn't match any of the main options, now try a wildcard
        for option in self.options:
            if option.hotkey == '*':
                return option.action()

    def __str__(self):
        return '\n'.join([str(option) for option in self.options])


class MenuFactory(object):
    def __init__(self, screen, writer, remover):
        self.default_options = []
        self.screen = screen
        self.writer = writer
        self.remover = remover

    def add_default_option(self, option):
        self.default_options.append(option)

    def new_menu(self, title):
        menu = Menu(title, self.screen,
                    self.writer,
                    self.remover)
        for opt in self.default_options:
            menu.add_option(opt)

        return menu



import json
import random
import simplejson
import time
import unittest
import urllib2

import tasksitter.main as main


class HTTPMonitoringTests(unittest.TestCase):

    def start_http_server(self, args):
        self.port = 1024 + int(10000 * random.random())
        print "Port %s Chosen" % self.port
        args.append("--http-monitoring")
        args.append("--http-monitoring-port=%s" % self.port)
        self.stats, self.httpd, self.harness = main.main(
            args, wait_for_child=False)

        # stop execution of this thread to give the HTTP thread time to
        # boot up
        time.sleep(.1)

    def stop_http_server(self):
        self.httpd.stop()
        self.harness.terminate_child()

    def make_call(self, url):
        data = None
        try:
            data = urllib2.urlopen(url).read()
        except:
            import traceback
            traceback.print_exc()
            self.fail()

        return data

    def run_check(self, args, path, output_format='json',
                  stop_server=True):
        self.start_http_server(args)
        data = self.make_call('http://localhost:%s/%s?format=%s&nohtml=1' % (
                self.port,
                path,
                output_format))

        if stop_server:
            self.stop_http_server()

        if not data:
            return {}

        if output_format == 'json':
            return simplejson.loads(data)
        else:
            rows = data.split("\n")
            return_data = {}
            for row in rows:
                if not row:
                    continue

                k, v = row.split('=', 1)
                return_data[k] = v

            return return_data

    def test_basic_monitoring_flat(self):
        data = self.run_check(['--cpu=.2',
                               '--command', 'sleep .2; ./spin.sh'],
                              path="stats",
                              output_format="flat")

        self.assertTrue("child_pid" in data)
        self.assertTrue(data["process_start_time"] is not None)
        self.assertTrue(data["task_start_time"] is not None)
        self.assertEquals(data["max_restarts"], '-1')
        self.assertEquals(data["num_task_starts"], '1')
        self.assertTrue("spin.sh" in data["command"])
        self.assertTrue("sleep" in data["command"])
        self.assertTrue("violated_CPU Constraint (0.2)" in data)

    def test_basic_monitoring_json(self):
        data = self.run_check(['--cpu=.2',
                               '--command', 'sleep .2; ./spin.sh'],
                              path="stats")

        self.assertTrue("child_pid" in data)
        self.assertTrue(data["process_start_time"] is not None)
        self.assertTrue(data["task_start_time"] is not None)
        self.assertEquals(data["max_restarts"], -1)
        self.assertEquals(data["num_task_starts"], 1)
        self.assertTrue("spin.sh" in data["command"])
        self.assertTrue("sleep" in data["command"])
        self.assertTrue("violated_CPU Constraint (0.2)" in data)

        self.assertTrue("file_version" in data)
        self.assertTrue("launch_location" in data)
        self.assertTrue("dir_version" in data)
        print data

    def test_logs_list_json(self):
        data = self.run_check(['--cpu=.2', "--ensure-alive",
                               '--command', 'id',
                               '--stdout-location', '/tmp/',
                               '--stderr-location', '/tmp/'],
                              path="logs",
                              output_format='json')

        self.assertTrue("stdout.0" in data)
        self.assertTrue("stderr.0" in data)
        self.assertTrue(len(data["stdout.0"]) > 0)
        self.assertTrue(len(data["stderr.0"]) > 0)

    def test_logs_list_flat(self):
        data = self.run_check(['--cpu=.2', "--ensure-alive",
                               '--command', 'id',
                               '--stdout-location', '/tmp/',
                               '--stderr-location', '/tmp/'],
                              path="logs",
                              output_format="flat")

        self.assertTrue("stdout.0" in data)
        self.assertTrue("stderr.0" in data)
        self.assertTrue(len(data["stdout.0"]) > 0)
        self.assertTrue(len(data["stderr.0"]) > 0)

    def test_download_log_file(self):
        data = self.run_check(['--cpu=.2', "--ensure-alive",
                               '--command', 'echo -n "hello"; sleep 10',
                               '--stdout-location', '/tmp/'],
                              path="logs",
                              output_format="json",
                              stop_server=False)

        filename = data["stdout.0"]['url']

        log_data = self.make_call(
            'http://localhost:%s%s&nohtml=1' % (
                self.port, filename))
        self.stop_http_server()

        self.assertEquals(log_data, "hello")

    def test_download_log_file_by_name(self):
        data = self.run_check(['--cpu=.2', "--ensure-alive",
                               '--command', 'echo -n "hello"; sleep 10',
                               '--stdout-location', '/tmp/'],
                              path="logs",
                              output_format="json",
                              stop_server=False)

        print data

        filename = data["stdout.0"]['url']

        log_data = self.make_call(
            'http://localhost:%s/logfile?logname=stdout.0&nohtml=1' % (
                self.port))
        self.stop_http_server()

        self.assertEquals(log_data, "hello")



import logging
from dynect_client import DynectDNSClient

from dnsprovider import DNSProvider

logger = logging.getLogger(__name__)


class Dynect(DNSProvider):
    def __init__(self, options):
        self.customername = options['customername']
        self.username = options['username']
        self.password = options['password']
        self.default_domain = options['default_domain']

        self.client = DynectDNSClient(self.customername,
                                      self.username,
                                      self.password,
                                      self.default_domain)

    def find_domainname(self, hostName, default):
        # NOTE!  This won't work for domains like .co.uk
        if not hostName:
            return default

        domain_parts = hostName.split('.')
        return ".".join(domain_parts[-2:])

    def add_record(self, data, hostName, type="A", TTL=300, domainName=None):
        domainName = self.find_domainname(hostName, domainName)

        try:
            ret = self.client.addRecord(data=data,
                                        type=type,
                                        hostName=hostName,
                                        TTL=TTL,
                                        domainName=domainName)
        except:
            import traceback
            logger.error(traceback.format_exc())
            logger.error(self.client.get_errors())
            return False

        if not ret:
            logger.warn("Couldn't add %s to %s: %s" % (
                    data,
                    hostName,
                    self.client.get_errors()))

        return ret

    def remove_record(self, data, hostName, type="A", domainName=None):
        domainName = self.find_domainname(hostName, domainName)

        try:
            ret = self.client.deleteRecord(data=data,
                                           type=type,
                                           hostName=hostName,
                                           domainName=domainName)
        except:
            import traceback
            logger.error(traceback.format_exc())
            logger.error(self.client.get_errors())
            return False

        if not ret:
            logger.warn("Couldn't delete %s from %s: %s" % (
                    data,
                    hostName,
                    self.client.get_errors()))

        return ret

    def get_records(self, hostName=None, type="A", domainName=None):
        domainName = self.find_domainname(hostName, domainName)

        logger.debug("Get Records for %s" % hostName)
        try:
            records = self.client.getRecords(hostName=hostName,
                                             type=type,
                                             domainName=domainName)
        except:
            import traceback
            logger.error(traceback.format_exc())
            logger.error(self.client.get_errors())
            return []

        if not records:
            return []

        if not hostName:
            return records

        def get_record(record):
            if r.get('record', '') == None:
                return ''

            return r.get('record', '')

        records = [r for r in records if hostName in get_record(r['record'])]

        if type == "*":
            return records

        return [r for r in records if r['type'] == type]



#!/usr/bin/python -u
"""
This is the main file of the clustersitter application.

It's job is to manage an entire fleet of machines, each running
machine sitters and task sitters.  It reaches out across
the cluster and monitors everything, and exposes several semantic
interfaces (all via HTTP).  There are NAGIOS type monitoring
endpoints as well as an API for interactions via a command line tool.
"""
import logging
import os
import sys

from clustersitter import ClusterSitter
from machineconfig import MachineConfig
from sittercommon import arg_parser


def parse_args(args):
    parser = arg_parser.ArgumentParser(
        description="Start the cluster sitter daemon")

    parser.add_argument("--daemon", dest="daemon",
                        default=False,
                        action="store_true",
                        help='Daemonize and split from launching shell')

    parser.add_argument("--login-user", dest="username",
                        default="ubuntu", help="User to login as")

    return parser.parse_args(args=args)


def daemonize():
    pid = os.fork()
    if pid > 0:
        sys.exit(0)

    os.setsid()
    os.chdir("/")

    pid = os.fork()
    if pid > 0:
        sys.exit(0)

    os.umask(0)
    sys.stdout.flush()
    sys.stderr.flush()
    sys.stdin.close()

    print "Sitter PID: %s" % os.getpid()


def main(sys_args=None):

    if not sys_args:
        sys_args = sys.argv[1:]

    args = parse_args(sys_args)

    settings_module = os.getenv('CLUSTERSITTER_SETTINGS',
                                'clustersitter.settings')
    try:
        __import__(settings_module)
        settings = sys.modules[settings_module]
    except:
        print "Couldn't find settings module %s" % settings_module
        print "Please set the CLUSTERSITTER_SETTINGS environmental " + \
            "variable an appropriate clustersitter settings module"
        sys.exit(1)

    launch_location = os.getcwd()

    if args.daemon:
        daemonize()

    logging.getLogger().setLevel(logging.INFO)
    logging.basicConfig(
        format='%(asctime)s %(name)s:%(levelname)s %(message)s')
    logging.getLogger().handlers[0].setLevel(logging.DEBUG)

    sitter = ClusterSitter(daemon=args.daemon,
                           provider_config=settings.provider_config,
                           dns_provider_config=settings.dns_provider_config,
                           keys=settings.keys, login_user=settings.login_user,
                           log_location=settings.log_location,
                           launch_location=launch_location)
    sitter.start()

    if False:
        # For testing.
        logging.getLogger().setLevel(logging.ERROR)
        localhost = MachineConfig("localhost", "localhost",
                                  1, 1)

        sitter.state.set_provider("localhost", None)
        sitter.add_machines([localhost], False)

    # wait forever
    os.system("tail -f -n 100 %s" % (sitter.logfiles[0]))
    while True:
        try:
            for index, name in enumerate(sitter.logfiles):
                print "%s. %s" % (index, name)
            num = raw_input("Chose a log to view: ")
            try:
                num = int(num)
                os.system("tail -f %s" % (sitter.logfiles[num]))
            except:
                pass
        except KeyboardInterrupt:
            print "Caught Control-C, exiting"
            os._exit(0)



from sittercommon.api import ClusterState


def get_help_string():
    return "List machines in clustersitter"


def get_command():
    return "listmachines"


def get_parser(parser):
    return parser


def run_command(clustersitter_url=None):

    state = ClusterState(clustersitter_url)
    state.reload()
    for machine in state.machines:
        print "%s (%s)" % (machine.hostname, machine.config.ip)
        for taskname, taskdata in machine.tasks.iteritems():
            running = "Not Running"
            if taskdata['running']:
                running = "Running    "
            print "  - %s %s (%s)" % (
                running, taskname, taskdata['command'],)

        print ""



"""
Determine external address using a few different services.
"""

import re
import socket
import time
import unittest
import urllib2


__all__ = ['ExternalAddress', 'SimpleAddressLookup']


class ExternalAddress(object):
    """
    Class used to look up public internet address. This is also the base class
    for specific implementations.
    """

    implementations = []

    def __init__(self, reverse_lookup, timeout):
        """
        Initialize the object.

        @param reverse_lookup Whether or not to resolve IP addresses to
            hostnames.
        @param timeout The timeout on API HTTP requests.
        """
        self.timeout = timeout
        self.reverse_lookup = reverse_lookup

    @classmethod
    def add_implementation(cls, implementation):
        """
        Add an implementation to the lookup.

        @param implementation A class implementing the lookup routine.
        """
        cls.implementations.append(implementation)

    @classmethod
    def lookup(cls, reverse_lookup=False, timeout=3):
        """
        Perform the external address lookup.

        @param reverse_lookup Whether or not to resolve IP addresses to
            hostnames.
        @param timeout The timeout on API HTTP requests.
        @return The external address.
        """
        address = None
        for impl in cls.implementations:
            obj = impl(reverse_lookup, timeout)
            address = obj.lookup_external_address()
            if address is not None:
                break
        return address

    @classmethod
    def benchmark(cls, iterations=10):
        """
        Benchmark the available implementations and reoder them according to
        speed. The Hostname implementation will always be last.

        @param iteration The number of times to execute each implementation
            during the benchmark.
        """
        has_hostname = False
        impls = []
        for impl in cls.implementations:
            if impl == Hostname:
                has_hostname = True
            else:
                now = time.time()
                obj = impl(False, 5)
                for i in range(iterations):
                    obj.lookup_external_address()
                impls.append((time.time() - now, impl))

        def cmp(a, b):
            return 1 if (a[0] - b[0]) > 0 else -1
        cls.implementations = [i[1] for i in sorted(impls, cmp)]
        if has_hostname:
            cls.implementations.append(Hostname)

    def read_uri(self, uri):
        """
        Get the contents of the page at the given URI. Return None on failure
        or timeout.

        @param uri The URI to retrieve the contents of.
        @return A string containing the contents of the requested page or None
            on failure or timeout.
        """
        try:
            return urllib2.urlopen(uri, timeout=self.timeout).read()
        except:
            return None

    def lookup_reverse_dns(self, ipaddr):
        """
        Do a reverse DNS lookup on an IP address. This does not obey the
        timeout value. If the reverse lookup fails then the value of ipaddr
        will be returned.

        @param ipaddr An IP address to do a reverse lookup on.
        @return The hostname of the IP or the original IP if that fails.
        """
        try:
            address = socket.gethostbyaddr(ipaddr)
            ipaddr = address[0]
        except:
            pass
        return ipaddr


class Hostname(ExternalAddress):
    """
    Return the hostname of the machine in the hopes that it's correct. This is
    used as a fallback method.
    """

    def lookup_external_address(self):
        return socket.getfqdn()


class SimpleAddressLookup(ExternalAddress):
    """
    Class implementing simple HTTP lookup with or without a regular expression.
    Extend this class and override the uri and regex attributes for a custom
    lookup implementation.

    Attributes:
        uri -- The URI of a page containing an the external IP.
        regex -- If not none should be a regex where group 1 contains the
            matched IP.
    """

    uri = None
    regex = None

    def lookup_external_address(self):
        address = self.read_uri(self.uri)
        if address is None:
            return None
        if self.regex is None:
            address = address.strip()
        else:
            m = re.search(self.regex, address, re.IGNORECASE)
            if m is None:
                return None
            address = m.group(1).strip()
        if self.reverse_lookup:
            address = self.lookup_reverse_dns(address)
        return address


class DynDNS(SimpleAddressLookup):
    """
    Do an address lookup using checkip.dyndns.org.
    """
    uri = 'http://checkip.dyndns.org/'
    regex = r'Address:([^<]*)<'


class ICanHazIP(SimpleAddressLookup):
    """
    Do an address lookup using icanhazip.com.
    """
    uri = 'http://icanhazip.com/'


class IfconfigDotMe(SimpleAddressLookup):
    """
    Do an address lookup using ifconfig.me.
    """
    uri = 'http://ifconfig.me/ip'


ExternalAddress.add_implementation(ICanHazIP)
ExternalAddress.add_implementation(DynDNS)
ExternalAddress.add_implementation(IfconfigDotMe)
ExternalAddress.add_implementation(Hostname)


class TestExternalAddress(unittest.TestCase):
    """
    Run tests against external address functionality.
    """

    def test_implementations(self):
        """
        Test lookup implementations. They should all work within five seconds
        and return the same external IP.
        """
        address = None
        for impl in ExternalAddress.implementations:
            if impl == Hostname:
                continue
            obj = impl(False, 5)
            if address is None:
                address = obj.lookup_external_address()
            else:
                lookup = obj.lookup_external_address()
                self.assertEqual(
                    address, lookup,
                    "invalid address returned by %s ('%s' != '%s')" %
                    (impl.__name__, address, lookup))

    def test_benchmark(self):
        """
        Verify that the implemenatations have been added in order of response
        time.
        """
        impls = list(ExternalAddress.implementations)
        ExternalAddress.benchmark()
        self.assertEqual(
            ExternalAddress.implementations, impls,
            "implemenations not in order of response time")

    def test_hostname(self):
        """
        Test the fallback Hostname implementation.
        """
        fqdn = socket.getfqdn()
        address = Hostname(False, 5).lookup_external_address()
        self.assertEqual(
            fqdn, address,
            "fallback Hostname implementation invalid ('%s' != '%s')" %
            (fqdn, address))

    def test_reverse_lookup(self):
        """
        Test reverse DNS lookup.
        """
        ipaddr = '8.8.8.8'
        expect = 'google-public-dns-a.google.com'
        address = ExternalAddress(False, 5).lookup_reverse_dns(ipaddr)
        self.assertEqual(
            expect, address,
            "invalid reverse lookup for ip %s ('%s' != '%s')" %
            (ipaddr, expect, address))


if __name__ == '__main__':
    ExternalAddress.lookup()



import json
import re
import os
import sys

CFG_FILE = os.path.expanduser("~/.cerebro.cfg")


class Output(object):
    def __init__(self):
        self.quiet = False

    def echo(self, msg):
        if not self.quiet:
            print msg

    def stderr(self, msg):
        if not self.quiet:
            sys.stderr.write(msg)

output = Output()


def strip_html(text):
    return re.sub('<[^<]+?>', '', text)


def load_defaults():
    default_options = {}

    try:
        default_data = open(CFG_FILE).read()
        default_options = json.loads(default_data)
    except:
        pass

    return default_options


def write_defaults(default_options):
    default_file = open(CFG_FILE, 'w')
    default_file.write(json.dumps(default_options))
    default_file.close()




import logging
from threading import Thread

try:
    from Queue import Empty, Full, Queue
    (Empty, Full, Queue)  # pyflakes fix
except ImportError:
    from queue import Empty, Full, Queue

logger = logging.getLogger(__name__)


class ClusterActionRunner(Thread):
    """
    Run a cluster action in a thread.
    """

    def __init__(self, action):
        """Initialize the runner."""
        self.thread = None
        self.action = action

    def is_running(self):
        """
        Check if the queue consumer is running.

        @return True if the consumer is running or False.
        """
        return self.thread is not None and self.thread.is_alive()

    def start(self):
        """
        Start the action thread.

        @return True if the thread was started or False if it was already
            running.
        """
        if self.is_running():
            return False

        def run():
            try:
                logger.info("action '%s' running" % self.action)
                self.action.run()
                logger.info("action '%s' complete" % self.action)
            except:
                import traceback
                logger.error("action '%s' failed" % self.action)
                logger.error(traceback.format_exc())
            finally:
                self.queue.task_done()

        self.thread = Thread(target=run, name=self.action)
        self.thread.start()
        return True

    def stop(self):
        """
        Signal the action to stop if supported. A method called stop() is
        checked for on the action and should return True on success. This
        method does not block.

        @return True if the action was signaled to stop or False if the thread
            was not running or the action does not support stopping.
        """
        if self.is_running() and hasattr(self.action, 'stop'):
            return self.action.stop()
        return False

    def join(self, timeout=None):
        """
        Block until the thread completes or the given timeout expires.

        @param timeout The number of seconds to wait for the thread to
            complete. Defaults to infinite.
        @return True if the thread has stopped or False if it is still running.
            If a timeout is not specified it will always be True.
        """
        if self.thread is not None:
            self.thread.join(timeout)
        return not self.is_running()


class MachineActionQueue(object):
    """
    Queue for running machine actions. Runs queues actions in a single thread
    until the queue empties.
    """

    timeout = 30

    def __init__(self, machine):
        """
        Initialize the queue.

        @param machine The machine this queue is operating on.
        """
        self.machine = machine
        self.name = "ActionQueue(%s)" % machine
        self.queue = Queue()
        self.stop = False
        self.thread = None

    def add(self, action):
        """
        Queue an action.

        @param action The action to queue.
        @return True on success or False if the action's machine does not match
        the queue's machine or if the queue add times out.
        """
        if self.machine == getattr(action, 'machine', None):
            try:
                self.queue.put(action, True, self.timeout)
                return True
            except Full:
                logger.error("%s: timed out while queuing" % action)
        return False

    def is_empty(self):
        """
        Check if the queue is empty.

        @return True if the queue is empty or False.
        """
        return self.queue.empty()

    def is_running(self):
        """
        Check if the queue consumer is running.

        @return True if the consumer is running or False.
        """
        return self.thread is not None and self.thread.is_alive()

    def start(self):
        """
        Start the queue consumer thread. The thread will only be run if the
        queue is non-empty. The thread will terminate once the queue is empty.

        @return True if the consumer thread was started or False if the thread
            is already running or the queue is empty.
        """
        if self.is_running() or self.is_empty():
            return False

        def run():
            logger.info("action queue consumer started")
            self.stop = False
            while not self.queue.empty() and not self.stop:
                try:
                    action = self.queue.get(True, self.timeout)
                except Empty:
                    logger.warn("queue timed out, retrying")
                    continue

                try:
                    logger.info("action '%s' running" % action)
                    action.run()
                    logger.info("action '%s' complete" % action)
                except:
                    import traceback
                    logger.error("action '%s' failed" % action)
                    logger.error(traceback.format_exc())
                finally:
                    self.queue.task_done()
            logger.info("action queue thread stopped")

        self.thread = Thread(target=run, name=str(self))
        self.thread.start()
        return True

    def stop(self):
        """
        Signal the queue consumer thread to stop. This method does not block.

        @return True if the consumer thread was signalled to stop or False if
            the thread is not running.
        """
        if not self.is_running():
            return False
        self.stop = True
        return True

    def join(self, timeout=None):
        """
        Block until the consumer thread completes or the given timeout expires.

        @param timeout The number of seconds to wait for the thread to
            complete. Defaults to infinite.
        @return True if the thread has stopped or False if it is still running.
            If a timeout is not specified it will always be True.
        """
        if self.thread is not None:
            self.thread.join(timeout)
        return not self.is_running()

    def __str__(self):
        """Return the name of the action queue."""
        return self.name

    def __repr__(self):
        """Return the name of the action queue."""
        return self.name


class ClusterActionManager(object):
    """
    Manage cluster actions. Each machine has a dedicated queue for actions
    which need to be performed. Each queue has a single consumer thread.
    Threads are cleaned up when their associated queues empty.
    """

    def __init__(self):
        """Initialize the manager."""
        self.pending = []
        self.queues = {}
        self.runners = []

    def add(self, action):
        """
        Add an action to the queue. The action is not actually queued until
        process() is called.
        """
        self.pending.append(action)

    def process(self):
        """
        Process pending actions. If the action is related to a machine it is
        queued in the associated machine queue. Otherwise the action is run
        immediately. This method also cleans up completed actions and empty
        machine queues.
        """
        while len(self.pending):
            action = self.pending.pop(0)
            if hasattr(action, 'machine'):
                queue = self.queues.get(action.machine)
                if not queue:
                    queue = MachineActionQueue(action.machine)
                    self.queues[action.machine] = queue
                if not queue.add(action):
                    logger.error("%s: failed to queue action" % action)
                if not queue.is_running():
                    queue.start()
            else:
                runner = ClusterActionRunner(action)
                result = runner.start()
                if result:
                    self.runners.append(runner)
                else:
                    logger.error("%: failed to start action" % action)

        for name, queue in dict(self.queues).iteritems():
            if not queue.is_running() and queue.is_empty():
                del self.queues[name]

        for runner in list(self.runners):
            if not runner.is_running():
                self.runners.remove(runner)

    def stop(self):
        """
        Signal all actions to stop.
        """
        for queue in self.queues.values():
            queue.stop()
        for runner in self.runners:
            runner.stop()

    def join(self, timeout=None):
        """
        Wait for all thread queues to complete.

        @param timeout The number of seconds to block waiting for a queue
            thread to join. This will be applied to each queue thread. Defaults
            to infinite.
        @return True if all threads are stopped or False if some threads
            continue to run.
        """
        running = True
        for queue in self.queues.values():
            running = running and queue.join(timeout)
        for runner in self.runners:
            running = running and runner.join(timeout)
        return running


class ClusterAction(object):
    """
    Base class for all actions. Inheriting classes must implement the run()
    method.
    """

    def __init__(self, sitter):
        """
        Initialize the action.

        @param sitter The cluster sitter object.
        """
        self.name = "%s" % self.__class__.__name__
        self.sitter = sitter
        self.state = sitter.state

    def run(self):
        """Run the action."""

    def __str__(self):
        """Return the name of the action."""
        return self.name

    def __repr__(self):
        """Return the name of the action."""
        return self.name


class MachineAction(ClusterAction):
    """
    Base class for action executed against a machine. Inheriting classes must
    implement the run() method.
    """

    def __init__(self, sitter, zone, machine):
        """
        Initialize the action.

        @param sitter The cluster sitter object.
        @param zone The zone the job is being deployed to.
        @param machine The machine to manage the job on.
        """
        super(MachineAction, self).__init__(sitter)
        self.machine = machine
        self.name = "%s(%s)" % (self.__class__.__name__, machine)
        self.zone = zone


class TaskAction(MachineAction):
    """
    Base class for managing a task on a machine. Inheriting classes must
    implement the run() method.
    """

    def __init__(self, sitter, zone, machine, task):
        """
        Initialize the action.

        @param sitter The cluster sitter object.
        @param zone The zone the job is being deployed to.
        @param machine The machine to manage the job on.
        @param task The task to run the action against.
        """
        super(TaskAction, self).__init__(sitter, zone, machine)
        self.name = "%s(%s, %s)" % (self.__class__.__name__, machine, task)
        self.task = task


class StartTaskAction(TaskAction):
    """Start a task."""

    def run(self):
        """Run the action."""
        job = self.state.get_job(self.task)
        if not job:
            logger.error("%s: job does not exist" % self)
        elif not self.machine.start_task(job):
            logger.error("%s: failed to start task" % self)


class RestartTaskAction(TaskAction):
    """Restart a task."""

    def run(self):
        """Run the action."""
        job = self.state.get_job(self.task)
        if not job:
            logger.error("%s: job does not exist" % self)
        elif not self.machine.restart_task(job):
            logger.error("%s: failed to restart task" % self)


class StopTaskAction(TaskAction):
    """Stop a task."""

    def run(self):
        """Run the action."""
        job = self.state.get_job(self.task)
        if not job:
            logger.error("%s: job does not exist" % self)
        elif not self.machine.stop_task(job):
            logger.error("%s: failed to stop task" % self)


class AddTaskAction(TaskAction):
    """Add a task to a machine."""

    def run(self):
        """Run the action."""
        job = self.state.get_job(self.task)
        if not job:
            logger.error("%s: job does not exist" % self)
        elif not job.deploy(self.zone, self.machine):
            logger.error("%s: failed to deploy task" % self)
        elif not self.state.start_task(self.machine, self.task):
            logger.error(
                "%s: failed to start task after deploying" % self)


class RemoveTaskAction(TaskAction):
    """Remove a task from a machine."""

    def run(self):
        """Run the action."""
        job = self.state.get_job(self.task)
        if not job:
            logger.error("%s: job does not exist" % self)
        elif not self.machine.stop_task(job):
            logger.error("%s: failed to remove task" % self)


class DeployMachineAction(ClusterAction):
    """Deploy a master job to a new machine."""

    def __init__(self, sitter, zone, job):
        """
        Initialize the action.

        @param sitter The cluster sitter object.
        @param zone The zone to deploy the tasks to.
        @param job The job to deploy.
        """
        super(DeployMachineAction, self).__init__(sitter)
        self.job = job
        self.name = "%s(%s, %s)" % (
            self.__class__.__name__, zone, job.name)
        self.stop = False
        self.zone = zone

    def run(self):
        """Run the action."""
        if not self.job.deploy(self.zone):
            logger.error(
                "%s: failed to deploy new machine" % self)
            return
        # Take our new machine out of Deploying mode.
        self.state.update_machine(self.machine, self.state.Active)


class RedeployMachineAction(MachineAction):
    """Redeploy an unreachable machine."""

    def __init__(self, sitter, zone, machine, job):
        """
        Redeploy a machinesitter job to a machine.

        @param sitter The cluster sitter object.
        @param zone The zone the job is being deployed to.
        @param machine The machine to manage the job on.
        @param job The repair job to run.
        """
        super(RedeployMachineAction, self).__init__(sitter, zone, machine)
        self.job = job

    def run(self):
        """Run the action."""
        logger.info("%s: redeploying machine sitter" % self)
        if self.job.deploy(self.zone, self.machine, True):
            self.state.update_machine(self.machine, self.state.Active)
            logger.info(
                "%s: successfully redeployed machine")
        else:
            logger.info(
                "%s: failed to redeploy, decomissioning and reassigning jobs" %
                self)
            self.state.detach_machine(self.machine)
            self.sitter.decomission_machine(self.machine)


class DecomissionMachineAction(MachineAction):
    """Decomission an existing machine."""

    def run(self):
        """Run the action."""
        self.sitter.decomission_machine(self.machine)



import os
import sys
import time

from clustersitter import MonitoredMachine, ProductionJob
from sittercommon.api import ClusterState
from sittercommon.utils import output


def get_help_string():
    return "Login to a cerebro managed machine"


def get_command():
    return "login"


def get_parser(parser):
    parser.add_argument("--remotecommand",
                        dest="remote_command",
                        help='Command to run on machine or a @file to run')

    parser.add_argument(dest="name",
                        nargs='*',
                        help='Job or machine to log in to')

    return parser


def get_options(state, name):
    if not name:
        # Show all jobs, then show machines
        return state.jobs + state.machines
    else:
        name = name.lower()
        algos = [
            lambda x, y: x == y,
            lambda x, y: x.startswith(y),
            lambda x, y: x in y]

        for algo in algos:
            for jobname in state.get_job_names():
                if algo(jobname.lower(), name):
                    output.stderr("\nUsing **%s** as a match for %s\n\n" % (
                        jobname, name))
                    return state.get_machines_for_job(state.get_job(jobname))

    return []


def login(state, machine, command=None):
    key = state.provider_config.get_key_for_zone(
        machine.config.shared_fate_zone)

    key_loc = state.find_key(key)
    login_user = state.login_user
    if not key_loc:
        output.stderr("Couldn't find key to login to %s" % machine)
        sys.exit(1)

    output.stderr("opening shell to %s...\n" % machine)
    options = ("-o LogLevel=quiet -o UserKnownHostsFile=/dev/null " +
               "-o StrictHostKeyChecking=no")
    remote_cmd = ''
    if command:
        remote_cmd = "'%s'" % command

    cmd = "ssh -t %s -i '%s' %s@%s %s" % (
        options, key_loc,
        login_user,
        machine.hostname,
        remote_cmd)

    # Try 3 times to login
    for i in xrange(3):
        output.stderr("%s\n" % cmd)
        start = time.time()
        ret = os.system(cmd)
        elapsed = time.time() - start
        if ret != 255 or elapsed > 30:
            return
        else:
            output.stderr("Login failed, trying again (%s/3)...\n" % (
                i + 1))


def run_command(clustersitter_url=None,
                name=None,
                remote_command=None):
    if remote_command and remote_command.startswith('@'):
        remote_command = open(remote_command[1:]).read()

    state = ClusterState(clustersitter_url)
    state.reload()
    if isinstance(name, list):
        name = ' '.join(name)

    def menu(options):
        if not options:
            output.stderr("No matches for %s found, trying all\n" % name)
            options = get_options(state, None)

        if len(options) == 1 and isinstance(options[0], MonitoredMachine):
            output.stderr("Only one machine found, logging in...\n")
            return login(state, options[0], remote_command)

        selected = None
        while not selected:
            for index, option in enumerate(options):
                output.echo("%s. %s" % (index, option))

            result = raw_input("Chose a machine or job to log into: ")
            try:
                selected = options[int(result)]
            except:
                continue

        if isinstance(selected, ProductionJob):
            return menu(get_options(state, selected.name))

        if isinstance(selected, MonitoredMachine):
            return login(state, selected, remote_command)

    options = get_options(state, name)
    menu(options)
